import asyncio
import threading

import bleak


class BleakAdapter:
    @staticmethod
    def scan_toys(timeout: float = 5.0):
        return asyncio.run(bleak.BleakScanner.discover(timeout))

    @staticmethod
    def scan_toy(name: str, timeout: float = 5.0):
        return asyncio.run(
            bleak.BleakScanner.find_device_by_filter(
                lambda _, a: a.local_name == name, timeout))

    def __init__(self, address):
        self.__event_loop = asyncio.new_event_loop()
        self.__device = bleak.BleakClient(address, timeout=5.0)
        self.__lock = threading.Lock()
        self.__thread = threading.Thread(target=self.__event_loop.run_forever)
        self.__thread.start()
        try:
            self.__execute(self.__device.connect())
        except:
            self.close(False)
            raise

    def __execute(self, coroutine):
        with self.__lock:
            return asyncio.run_coroutine_threadsafe(coroutine, self.__event_loop).result()

    def close(self, disconnect=True):
        if disconnect:
            self.__execute(self.__device.disconnect())
        with self.__lock:
            self.__event_loop.call_soon_threadsafe(self.__event_loop.stop)
            self.__thread.join()
        self.__event_loop.close()

    def set_callback(self, uuid, cb):
        self.__execute(self.__device.start_notify(uuid, cb))

    def write(self, uuid, data):
        self.__execute(self.__device.write_gatt_char(uuid, data, True))



from enum import IntEnum

from spherov2.toy.r2d2 import R2D2
from spherov2.types import ToyType


class R2Q5(R2D2):
    toy_type = ToyType('R2-Q5', 'Q5-', 'Q5', .12)

    class Animations(IntEnum):
        CHARGER_1 = 0
        CHARGER_2 = 1
        CHARGER_3 = 2
        CHARGER_4 = 3
        CHARGER_5 = 4
        CHARGER_6 = 5
        CHARGER_7 = 6
        EMOTE_ALARM = 7
        EMOTE_ANGRY = 8
        EMOTE_ATTENTION = 9
        EMOTE_FRUSTRATED = 10
        EMOTE_DRIVE = 11
        EMOTE_EXCITED = 12
        EMOTE_SEARCH = 13
        EMOTE_SHORT_CIRCUIT = 14
        EMOTE_LAUGH = 15
        EMOTE_NO = 16
        EMOTE_RETREAT = 17
        EMOTE_FIERY = 18
        EMOTE_UNDERSTOOD = 19
        EMOTE_YES = 21
        EMOTE_SCAN = 22
        EMOTE_SURPRISED = 24
        IDLE_1 = 25
        IDLE_2 = 26
        IDLE_3 = 27
        WWM_ANGRY = 31
        WWM_ANXIOUS = 32
        WWM_BOW = 33
        WWM_CONCERN = 34
        WWM_CURIOUS = 35
        WWM_DOUBLE_TAKE = 36
        WWM_EXCITED = 37
        WWM_FIERY = 38
        WMM_FRUSTRATED = 39
        WWM_HAPPY = 40
        WWM_JITTERY = 41
        WWM_LAUGH = 42
        WWM_LONG_SHAKE = 43
        WWM_NO = 44
        WWM_OMINOUS = 45
        WWM_RELIEVED = 46
        WWM_SAD = 47
        WWM_SCARED = 48
        WWM_SHAKE = 49
        WWM_SURPRISED = 50
        WWM_TAUNTING = 51
        WWM_WHISPER = 52
        WWM_YELLING = 53
        WWM_YOOHOO = 54



from spherov2.commands import Commands


class SystemMode(Commands):
    _did = 18

    @staticmethod
    def enable_desktoy_mode(toy, enable: bool, proc=None):
        toy._execute(SystemMode.__encode(toy, 41, proc, [int(enable)]))

    @staticmethod
    def get_out_of_box_state(toy, proc=None):
        return bool(toy._execute(SystemMode._encode(toy, 43, proc)).data[0])

    @staticmethod
    def enable_out_of_box_state(toy, enable: bool, proc=None):
        toy._execute(SystemMode.__encode(toy, 44, proc, [int(enable)]))



from enum import IntEnum
from functools import lru_cache, partialmethod

from spherov2.commands.api_and_shell import ApiAndShell
from spherov2.commands.connection import Connection
from spherov2.commands.drive import Drive
from spherov2.commands.factory_test import FactoryTest
from spherov2.commands.firmware import Firmware
from spherov2.commands.io import IO
from spherov2.commands.power import Power
from spherov2.commands.sensor import Sensor
from spherov2.commands.system_info import SystemInfo
from spherov2.commands.system_mode import SystemMode
from spherov2.controls.v2 import DriveControl, LedControl, StreamingControl, Processors
from spherov2.toy import Toy, ToyV2
from spherov2.types import ToyType


class RVR(ToyV2):
    toy_type = ToyType('Sphero RVR', 'RV-', 'RV', .075)
    _require_target = True

    class LEDs(IntEnum):
        RIGHT_HEADLIGHT_RED = 0
        RIGHT_HEADLIGHT_GREEN = 1
        RIGHT_HEADLIGHT_BLUE = 2
        LEFT_HEADLIGHT_RED = 3
        LEFT_HEADLIGHT_GREEN = 4
        LEFT_HEADLIGHT_BLUE = 5
        LEFT_STATUS_INDICATION_RED = 6
        LEFT_STATUS_INDICATION_GREEN = 7
        LEFT_STATUS_INDICATION_BLUE = 8
        RIGHT_STATUS_INDICATION_RED = 9
        RIGHT_STATUS_INDICATION_GREEN = 10
        RIGHT_STATUS_INDICATION_BLUE = 11
        BATTERY_DOOR_REAR_RED = 12
        BATTERY_DOOR_REAR_GREEN = 13
        BATTERY_DOOR_REAR_BLUE = 14
        BATTERY_DOOR_FRONT_RED = 15
        BATTERY_DOOR_FRONT_GREEN = 16
        BATTERY_DOOR_FRONT_BLUE = 17
        POWER_BUTTON_FRONT_RED = 18
        POWER_BUTTON_FRONT_GREEN = 19
        POWER_BUTTON_FRONT_BLUE = 20
        POWER_BUTTON_REAR_RED = 21
        POWER_BUTTON_REAR_GREEN = 22
        POWER_BUTTON_REAR_BLUE = 23
        LEFT_BRAKELIGHT_RED = 24
        LEFT_BRAKELIGHT_GREEN = 25
        LEFT_BRAKELIGHT_BLUE = 26
        RIGHT_BRAKELIGHT_RED = 27
        RIGHT_BRAKELIGHT_GREEN = 28
        RIGHT_BRAKELIGHT_BLUE = 29
        UNDERCARRIAGE_WHITE = 30

    ping = ApiAndShell.ping
    get_api_protocol_version = ApiAndShell.get_api_protocol_version
    send_command_to_shell = ApiAndShell.send_command_to_shell
    add_send_string_to_console_listener = partialmethod(Toy._add_listener, ApiAndShell.send_string_to_console)
    remove_send_string_to_console_listener = partialmethod(Toy._remove_listener, ApiAndShell.send_string_to_console)
    get_supported_dids = ApiAndShell.get_supported_dids
    get_supported_cids = ApiAndShell.get_supported_cids

    get_main_app_version = SystemInfo.get_main_app_version
    get_bootloader_version = SystemInfo.get_bootloader_version
    get_board_revision = partialmethod(SystemInfo.get_board_revision, proc=Processors.PRIMARY)
    get_mac_address = partialmethod(SystemInfo.get_mac_address, proc=Processors.PRIMARY)
    get_stats_id = partialmethod(SystemInfo.get_stats_id, proc=Processors.PRIMARY)
    get_processor_name = SystemInfo.get_processor_name
    get_boot_reason = partialmethod(SystemInfo.get_boot_reason, proc=Processors.PRIMARY)
    get_last_error_info = partialmethod(SystemInfo.get_last_error_info, proc=Processors.PRIMARY)
    write_config_block = partialmethod(SystemInfo.write_config_block, proc=Processors.PRIMARY)
    get_config_block = partialmethod(SystemInfo.get_config_block, proc=Processors.PRIMARY)
    set_config_block = partialmethod(SystemInfo.set_config_block, proc=Processors.PRIMARY)
    erase_config_block = partialmethod(SystemInfo.erase_config_block, proc=Processors.PRIMARY)
    get_swd_locking_status = SystemInfo.get_swd_locking_status
    get_manufacturing_date = partialmethod(SystemInfo.get_manufacturing_date, proc=Processors.PRIMARY)
    get_sku = partialmethod(SystemInfo.get_sku, proc=Processors.PRIMARY)
    get_core_up_time_in_milliseconds = partialmethod(SystemInfo.get_core_up_time_in_milliseconds,
                                                     proc=Processors.PRIMARY)
    get_event_log_status = SystemInfo.get_event_log_status
    get_event_log_data = SystemInfo.get_event_log_data
    clear_event_log = SystemInfo.clear_event_log
    enable_sos_message_notify = partialmethod(SystemInfo.enable_sos_message_notify, proc=Processors.PRIMARY)
    add_sos_message_notify_listener = partialmethod(Toy._add_listener, SystemInfo.sos_message_notify)
    remove_sos_message_notify_listener = partialmethod(Toy._remove_listener, SystemInfo.sos_message_notify)
    get_sos_message = partialmethod(SystemInfo.get_sos_message, proc=Processors.PRIMARY)
    clear_sos_message = partialmethod(SystemInfo.clear_sos_message, proc=Processors.PRIMARY)

    get_out_of_box_state = partialmethod(SystemMode.get_out_of_box_state, proc=Processors.PRIMARY)
    enable_out_of_box_state = partialmethod(SystemMode.enable_out_of_box_state, proc=Processors.PRIMARY)

    enter_deep_sleep = partialmethod(Power.enter_deep_sleep, proc=Processors.PRIMARY)
    sleep = partialmethod(Power.sleep, proc=Processors.PRIMARY)
    force_battery_refresh = partialmethod(Power.force_battery_refresh, proc=Processors.PRIMARY)
    wake = partialmethod(Power.wake, proc=Processors.PRIMARY)
    get_battery_percentage = partialmethod(Power.get_battery_percentage, proc=Processors.PRIMARY)
    get_battery_voltage_state = partialmethod(Power.get_battery_voltage_state, proc=Processors.PRIMARY)
    add_will_sleep_notify_listener = partialmethod(Toy._add_listener, Power.will_sleep_notify)
    remove_will_sleep_notify_listener = partialmethod(Toy._remove_listener, Power.will_sleep_notify)
    add_did_sleep_notify_listener = partialmethod(Toy._add_listener, Power.did_sleep_notify)
    remove_did_sleep_notify_listener = partialmethod(Toy._remove_listener, Power.did_sleep_notify)
    enable_battery_voltage_state_change_notify = partialmethod(Power.enable_battery_voltage_state_change_notify,
                                                               proc=Processors.PRIMARY)
    add_battery_voltage_state_change_notify_listener = partialmethod(Toy._add_listener,
                                                                     Power.battery_voltage_state_change_notify)
    remove_battery_voltage_state_change_notify_listener = partialmethod(Toy._remove_listener,
                                                                        Power.battery_voltage_state_change_notify)
    get_battery_voltage_in_volts = partialmethod(Power.get_battery_voltage_in_volts, proc=Processors.PRIMARY)
    get_battery_voltage_state_thresholds = partialmethod(Power.get_battery_voltage_state_thresholds,
                                                         proc=Processors.PRIMARY)
    get_current_sense_amplifier_current = partialmethod(Power.get_current_sense_amplifier_current,
                                                        proc=Processors.PRIMARY)
    get_efuse_fault_status = partialmethod(Power.get_efuse_fault_status, proc=Processors.PRIMARY)
    add_efuse_fault_occurred_notify_listener = partialmethod(Toy._add_listener, Power.efuse_fault_occurred_notify)
    remove_efuse_fault_occurred_notify_listener = partialmethod(Toy._remove_listener, Power.efuse_fault_occurred_notify)
    enable_efuse = partialmethod(Power.enable_efuse, proc=Processors.PRIMARY)

    set_raw_motors = partialmethod(Drive.set_raw_motors, proc=Processors.SECONDARY)
    reset_yaw = partialmethod(Drive.reset_yaw, proc=Processors.SECONDARY)
    drive_with_heading = partialmethod(Drive.drive_with_heading, proc=Processors.SECONDARY)
    set_control_system_type = partialmethod(Drive.set_control_system_type, proc=Processors.SECONDARY)
    set_component_parameters = partialmethod(Drive.set_component_parameters, proc=Processors.SECONDARY)
    get_component_parameters = partialmethod(Drive.get_component_parameters, proc=Processors.SECONDARY)
    set_custom_control_system_timeout = partialmethod(Drive.set_custom_control_system_timeout,
                                                      proc=Processors.SECONDARY)
    enable_motor_stall_notify = partialmethod(Drive.enable_motor_stall_notify, proc=Processors.SECONDARY)
    add_motor_stall_notify_listener = partialmethod(Toy._add_listener, Drive.motor_stall_notify)
    remove_motor_stall_notify_listener = partialmethod(Toy._remove_listener, Drive.motor_stall_notify)
    enable_motor_fault_notify = partialmethod(Drive.enable_motor_fault_notify, proc=Processors.SECONDARY)
    add_motor_fault_notify_listener = partialmethod(Toy._add_listener, Drive.motor_fault_notify)
    remove_motor_fault_notify_listener = partialmethod(Toy._remove_listener, Drive.motor_fault_notify)
    get_motor_fault_state = partialmethod(Drive.get_motor_fault_state, proc=Processors.SECONDARY)

    enable_gyro_max_notify = partialmethod(Sensor.enable_gyro_max_notify, proc=Processors.SECONDARY)
    add_gyro_max_notify_listener = partialmethod(Toy._add_listener, Sensor.gyro_max_notify)
    remove_gyro_max_notify_listener = partialmethod(Toy._remove_listener, Sensor.gyro_max_notify)
    reset_locator_x_and_y = partialmethod(Sensor.reset_locator_x_and_y, proc=Processors.SECONDARY)
    set_locator_flags = partialmethod(Sensor.set_locator_flags, proc=Processors.SECONDARY)
    get_bot_to_bot_infrared_readings = partialmethod(Sensor.get_bot_to_bot_infrared_readings, proc=Processors.SECONDARY)
    get_rgbc_sensor_values = partialmethod(Sensor.get_rgbc_sensor_values, proc=Processors.PRIMARY)
    magnetometer_calibrate_to_north = partialmethod(Sensor.magnetometer_calibrate_to_north, proc=Processors.SECONDARY)
    add_magnetometer_north_yaw_notify_listener = partialmethod(Toy._add_listener, Sensor.magnetometer_north_yaw_notify)
    remove_magnetometer_north_yaw_notify_listener = partialmethod(Toy._remove_listener,
                                                                  Sensor.magnetometer_north_yaw_notify)
    start_robot_to_robot_infrared_broadcasting = partialmethod(Sensor.start_robot_to_robot_infrared_broadcasting,
                                                               proc=Processors.SECONDARY)
    start_robot_to_robot_infrared_following = partialmethod(Sensor.start_robot_to_robot_infrared_following,
                                                            proc=Processors.SECONDARY)
    stop_robot_to_robot_infrared_broadcasting = partialmethod(Sensor.stop_robot_to_robot_infrared_broadcasting,
                                                              proc=Processors.SECONDARY)
    add_robot_to_robot_infrared_message_received_notify_listener = partialmethod(
        Toy._add_listener, Sensor.robot_to_robot_infrared_message_received_notify)
    remove_robot_to_robot_infrared_message_received_notify_listener = partialmethod(
        Toy._remove_listener, Sensor.robot_to_robot_infrared_message_received_notify)
    get_ambient_light_sensor_value = partialmethod(Sensor.get_ambient_light_sensor_value, proc=Processors.PRIMARY)
    stop_robot_to_robot_infrared_following = partialmethod(Sensor.stop_robot_to_robot_infrared_following,
                                                           proc=Processors.SECONDARY)
    start_robot_to_robot_infrared_evading = partialmethod(Sensor.start_robot_to_robot_infrared_evading,
                                                          proc=Processors.SECONDARY)
    stop_robot_to_robot_infrared_evading = partialmethod(Sensor.stop_robot_to_robot_infrared_evading,
                                                         proc=Processors.SECONDARY)
    enable_color_detection_notify = partialmethod(Sensor.enable_color_detection_notify, proc=Processors.PRIMARY)
    add_color_detection_notify_listener = partialmethod(Toy._add_listener, Sensor.color_detection_notify)
    remove_color_detection_notify_listener = partialmethod(Toy._remove_listener, Sensor.color_detection_notify)
    get_current_detected_color_reading = partialmethod(Sensor.get_current_detected_color_reading,
                                                       proc=Processors.PRIMARY)
    enable_color_detection = partialmethod(Sensor.enable_color_detection, proc=Processors.PRIMARY)
    configure_streaming_service = Sensor.configure_streaming_service
    start_streaming_service = Sensor.start_streaming_service
    stop_streaming_service = Sensor.stop_streaming_service
    clear_streaming_service = Sensor.clear_streaming_service
    add_streaming_service_data_notify_listener = partialmethod(Toy._add_listener, Sensor.streaming_service_data_notify)
    remove_streaming_service_data_notify_listener = partialmethod(Toy._remove_listener,
                                                                  Sensor.streaming_service_data_notify)
    enable_robot_infrared_message_notify = partialmethod(Sensor.enable_robot_infrared_message_notify,
                                                         proc=Processors.SECONDARY)
    send_infrared_message = partialmethod(Sensor.send_infrared_message, proc=Processors.SECONDARY)
    add_motor_current_notify_listener = partialmethod(Toy._add_listener, Sensor.motor_current_notify)
    remove_motor_current_notify_listener = partialmethod(Toy._remove_listener, Sensor.motor_current_notify)
    enable_motor_current_notify = partialmethod(Sensor.enable_motor_current_notify, proc=Processors.SECONDARY)
    get_motor_temperature = partialmethod(Sensor.get_motor_temperature, proc=Processors.SECONDARY)
    configure_sensitivity_based_collision_detection = partialmethod(
        Sensor.configure_sensitivity_based_collision_detection, proc=Processors.SECONDARY)
    enable_sensitivity_based_collision_detection_notify = partialmethod(
        Sensor.enable_sensitivity_based_collision_detection_notify, proc=Processors.SECONDARY)
    add_sensitivity_based_collision_detected_notify_listener = partialmethod(
        Toy._add_listener, Sensor.sensitivity_based_collision_detected_notify)
    remove_sensitivity_based_collision_detected_notify_listener = partialmethod(
        Toy._remove_listener, Sensor.sensitivity_based_collision_detected_notify)
    get_motor_thermal_protection_status = partialmethod(Sensor.get_motor_thermal_protection_status,
                                                        proc=Processors.SECONDARY)
    enable_motor_thermal_protection_status_notify = partialmethod(Sensor.enable_motor_thermal_protection_status_notify,
                                                                  proc=Processors.SECONDARY)
    add_motor_thermal_protection_status_notify_listener = partialmethod(Toy._add_listener,
                                                                        Sensor.motor_thermal_protection_status_notify)
    remove_motor_thermal_protection_status_notify_listener = partialmethod(
        Toy._remove_listener, Sensor.motor_thermal_protection_status_notify)

    set_bluetooth_name = partialmethod(Connection.set_bluetooth_name, proc=Processors.PRIMARY)
    get_bluetooth_name = partialmethod(Connection.get_bluetooth_name, proc=Processors.PRIMARY)
    get_bluetooth_advertising_name = partialmethod(Connection.get_bluetooth_advertising_name, proc=Processors.PRIMARY)

    set_all_leds_with_32_bit_mask = partialmethod(IO.set_all_leds_with_32_bit_mask, proc=Processors.PRIMARY)
    set_compressed_frame_player_one_color = partialmethod(IO.set_compressed_frame_player_one_color,
                                                          proc=Processors.PRIMARY)
    save_compressed_frame_player_animation = partialmethod(IO.save_compressed_frame_player_animation,
                                                           proc=Processors.PRIMARY)
    play_compressed_frame_player_animation = partialmethod(IO.play_compressed_frame_player_animation,
                                                           proc=Processors.PRIMARY)
    play_compressed_frame_player_frame = partialmethod(IO.play_compressed_frame_player_frame,
                                                       proc=Processors.PRIMARY)
    get_compressed_frame_player_list_of_frames = partialmethod(IO.get_compressed_frame_player_list_of_frames,
                                                               proc=Processors.PRIMARY)
    delete_all_compressed_frame_player_animations_and_frames = partialmethod(
        IO.delete_all_compressed_frame_player_animations_and_frames, proc=Processors.PRIMARY)
    pause_compressed_frame_player_animation = partialmethod(IO.pause_compressed_frame_player_animation,
                                                            proc=Processors.PRIMARY)
    resume_compressed_frame_player_animation = partialmethod(IO.resume_compressed_frame_player_animation,
                                                             proc=Processors.PRIMARY)
    reset_compressed_frame_player_animation = partialmethod(IO.reset_compressed_frame_player_animation,
                                                            proc=Processors.PRIMARY)
    add_compressed_frame_player_animation_complete_notify_listener = partialmethod(
        Toy._add_listener, IO.compressed_frame_player_animation_complete_notify)
    remove_compressed_frame_player_animation_complete_notify_listener = partialmethod(
        Toy._remove_listener, IO.compressed_frame_player_animation_complete_notify)
    assign_compressed_frame_player_frames_to_animation = partialmethod(
        IO.assign_compressed_frame_player_frames_to_animation, proc=Processors.PRIMARY)
    save_compressed_frame_player_animation_without_frames = partialmethod(
        IO.save_compressed_frame_player_animation_without_frames, proc=Processors.PRIMARY)
    play_compressed_frame_player_animation_with_loop_option = partialmethod(
        IO.play_compressed_frame_player_animation_with_loop_option, proc=Processors.PRIMARY)
    get_active_color_palette = partialmethod(IO.get_active_color_palette, proc=Processors.PRIMARY)
    set_active_color_palette = partialmethod(IO.set_active_color_palette, proc=Processors.PRIMARY)
    get_color_identification_report = partialmethod(IO.get_color_identification_report, proc=Processors.PRIMARY)
    load_color_palette = partialmethod(IO.load_color_palette, proc=Processors.PRIMARY)
    save_color_palette = partialmethod(IO.save_color_palette, proc=Processors.PRIMARY)
    get_compressed_frame_player_frame_info_type = partialmethod(IO.get_compressed_frame_player_frame_info_type,
                                                                proc=Processors.PRIMARY)
    save_compressed_frame_player16_bit_frame = partialmethod(IO.save_compressed_frame_player16_bit_frame,
                                                             proc=Processors.PRIMARY)
    release_led_requests = partialmethod(IO.release_led_requests, proc=Processors.PRIMARY)

    get_current_application_id = Firmware.get_current_application_id
    get_all_updatable_processors = partialmethod(Firmware.get_all_updatable_processors, proc=Processors.PRIMARY)
    get_version_for_updatable_processors = partialmethod(Firmware.get_version_for_updatable_processors,
                                                         proc=Processors.PRIMARY)
    set_pending_update_for_processors = partialmethod(Firmware.set_pending_update_for_processors,
                                                      proc=Processors.PRIMARY)
    get_pending_update_for_processors = partialmethod(Firmware.get_pending_update_for_processors,
                                                      proc=Processors.PRIMARY)
    reset_with_parameters = partialmethod(Firmware.reset_with_parameters, proc=Processors.PRIMARY)
    clear_pending_update_processors = partialmethod(Firmware.clear_pending_update_processors, proc=Processors.PRIMARY)
    get_factory_mode_challenge = FactoryTest.get_factory_mode_challenge
    enter_factory_mode = FactoryTest.enter_factory_mode
    exit_factory_mode = FactoryTest.exit_factory_mode
    get_chassis_id = FactoryTest.get_chassis_id
    enable_extended_life_test = partialmethod(FactoryTest.enable_extended_life_test, proc=Processors.PRIMARY)
    get_factory_mode_status = FactoryTest.get_factory_mode_status

    @property
    @lru_cache(None)
    def drive_control(self):
        return DriveControl(self)

    @property
    @lru_cache(None)
    def multi_led_control(self):
        return LedControl(self)

    @property
    @lru_cache(None)
    def sensor_control(self):
        return StreamingControl(self)



import importlib
from functools import partial
from typing import Iterable, List, Type, Callable

from spherov2.commands.sphero import Sphero
from spherov2.toy import Toy
from spherov2.toy.bb8 import BB8
from spherov2.toy.bb9e import BB9E
from spherov2.toy.bolt import BOLT
from spherov2.toy.mini import Mini
from spherov2.toy.ollie import Ollie
from spherov2.toy.r2d2 import R2D2
from spherov2.toy.r2q5 import R2Q5
from spherov2.toy.rvr import RVR


class ToyNotFoundError(Exception):
    ...


def all_toys(cls=Toy):
    subtypes = cls.__subclasses__()
    yield cls
    for sub in subtypes:
        yield from all_toys(sub)


def find_toys(*, timeout=5.0, toy_types: Iterable[Type[Toy]] = None,
              toy_names: Iterable[str] = None, adapter=None) -> List[Toy]:
    """Find toys that matches the criteria given.

    :param timeout: Device scanning timeout, in seconds.
    :param toy_types: Iterable of toy types (subclasses of :class:`Toy`) that needs to be scanned. Set to ``None`` to scan
                      all toy types available.
    :param toy_names: Iterable of strings of toy names that needs to be scanned. Set to ``None`` to scan toys with all
                      kinds of names.
    :param adapter: Kind of adapter to use for scanning bluetooth devices. Set to ``None`` to use default
                    :class:`BleakAdapter`.
    :return: A list of toys that are scanned.
    """
    if adapter is None:
        adapter = importlib.import_module(
            'spherov2.adapter.bleak_adapter').BleakAdapter
    if toy_names is not None:
        toy_names = set(toy_names)
    if toy_names is not None and len(toy_names) == 1:
        toy = adapter.scan_toy(list(toy_names)[0], timeout)
        if toy is None:
            return []
        toys = [toy]
    else:
        toys = adapter.scan_toys(timeout)
    if toy_types is None:
        toy_types = all_toys()
    ret = []
    for toy in toys:
        if toy.name is None:
            continue
        if toy_names is not None and toy.name not in toy_names:
            continue
        for toy_cls in toy_types:
            toy_type = toy_cls.toy_type
            if toy.name.startswith(toy_type.filter_prefix) and \
                    (toy_type.prefix is None or toy.name.startswith(toy_type.prefix)):
                ret.append(toy_cls(toy, adapter))
                break
    return ret


def find_toy(*, toy_name: str = None, **kwargs) -> Toy:
    """Find a single toy that matches the criteria given.

    :param toy_name: A string of toy name that needs to be scanned. Set to ``None`` to scan toy with all kinds of names.
    :param timeout: Device scanning timeout, in seconds.
    :param toy_types: List of toy types (subclasses of :class:`Toy`) that needs to be scanned. Set to ``None`` to scan
                      all toy types available.
    :param adapter: Kind of adapter to use for scanning bluetooth devices. Set to ``None`` to use default
                    :class:`BleakAdapter`.
    :return: A toy that is scanned.
    :raise ToyNotFoundError: If no toys could be found
    """
    toys = find_toys(toy_names=[toy_name] if toy_name else None, **kwargs)
    if not toys:
        raise ToyNotFoundError
    return toys[0]


find_Sphero: Callable[..., Sphero] = partial(find_toy, toy_types=[Sphero])
find_Ollie: Callable[..., Ollie] = partial(find_toy, toy_types=[Ollie])
find_Mini: Callable[..., Mini] = partial(find_toy, toy_types=[Mini])
find_BB8: Callable[..., BB8] = partial(find_toy, toy_types=[BB8])
find_BB9E: Callable[..., BB9E] = partial(find_toy, toy_types=[BB9E])
find_R2D2: Callable[..., R2D2] = partial(find_toy, toy_types=[R2D2])
find_R2Q5: Callable[..., R2Q5] = partial(find_toy, toy_types=[R2Q5])
find_RVR: Callable[..., RVR] = partial(find_toy, toy_types=[RVR])
find_BOLT: Callable[..., BOLT] = partial(find_toy, toy_types=[BOLT])



from spherov2.commands import Commands
from spherov2.helper import to_bytes, to_int


class FactoryTest(Commands):
    _did = 31

    @staticmethod
    def get_factory_mode_challenge(toy, proc=None):
        return to_int(toy._execute(FactoryTest._encode(toy, 19, proc)).data)

    @staticmethod
    def enter_factory_mode(toy, challenge: int, proc=None):
        toy._execute(FactoryTest._encode(toy, 20, proc, to_bytes(challenge, 4)))

    @staticmethod
    def exit_factory_mode(toy, proc=None):
        toy._execute(FactoryTest._encode(toy, 21, proc))

    @staticmethod
    def get_chassis_id(toy, proc=None):
        return to_int(toy._execute(FactoryTest._encode(toy, 39, proc)).data)

    @staticmethod
    def enable_extended_life_test(toy, enable, proc=None):
        toy._execute(FactoryTest._encode(toy, 49, proc, [int(enable)]))

    @staticmethod
    def get_factory_mode_status(toy, proc=None):
        return bool(toy._execute(FactoryTest._encode(toy, 52, proc)).data[0])



import threading
import time
from collections import OrderedDict, defaultdict
from concurrent import futures
from functools import partial
from queue import SimpleQueue
from typing import NamedTuple, Callable

from spherov2.controls.v1 import Packet as PacketV1
from spherov2.controls.v2 import Packet as PacketV2
from spherov2.types import ToyType


class ToySensor(NamedTuple):
    bit: int
    min_value: float
    max_value: float
    modifier: Callable[[float], float] = None


class Toy:
    toy_type = ToyType('Robot', None, 'Sphero', .06)
    sensors = OrderedDict()
    extended_sensors = OrderedDict()

    _send_uuid = '22bb746f-2ba1-7554-2d6f-726568705327'
    _response_uuid = '22bb746f-2ba6-7554-2d6f-726568705327'
    _handshake = [('22bb746f-2bbd-7554-2d6f-726568705327', bytearray(b'011i3')),
                  ('22bb746f-2bb2-7554-2d6f-726568705327', bytearray([7]))]
    _packet = PacketV1
    _require_target = False

    def __init__(self, toy, adapter_cls):
        self.address = toy.address
        self.name = toy.name

        self.__adapter = None
        self.__adapter_cls = adapter_cls
        self._packet_manager = self._packet.Manager()
        self.__decoder = self._packet.Collector(self.__new_packet)
        self.__waiting = defaultdict(SimpleQueue)
        self.__listeners = defaultdict(dict)
        self._sensor_controller = None

        self.__thread = None
        self.__packet_queue = SimpleQueue()

    def __repr__(self):
        return f'{self.name} ({self.address})'

    def __enter__(self):
        if self.__adapter is not None:
            raise RuntimeError('Toy already in context manager')
        self.__adapter = self.__adapter_cls(self.address)
        self.__thread = threading.Thread(target=self.__process_packet)
        try:
            for uuid, data in self._handshake:
                self.__adapter.write(uuid, data)
            self.__adapter.set_callback(self._response_uuid, self.__api_read)
            self.__thread.start()
        except:
            self.__exit__(None, None, None)
            raise
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__adapter.close()
        self.__adapter = None
        if self.__thread.is_alive():
            self.__packet_queue.put(None)
            self.__thread.join()
        self.__packet_queue = SimpleQueue()

    def __process_packet(self):
        while self.__adapter is not None:
            payload = self.__packet_queue.get()
            if payload is None:
                break
            # print('request ' + ' '.join([hex(c) for c in payload]))
            while payload:
                self.__adapter.write(self._send_uuid, payload[:20])
                payload = payload[20:]
            time.sleep(self.toy_type.cmd_safe_interval)

    def _execute(self, packet):
        if self.__adapter is None:
            raise RuntimeError('Use toys in context manager')
        self.__packet_queue.put(packet.build())
        return self._wait_packet(packet.id)

    def _wait_packet(self, key, timeout=10.0, check_error=False):
        future = futures.Future()
        self.__waiting[key].put(future)
        packet = future.result(timeout)
        if check_error:
            packet.check_error()
        return packet

    def _add_listener(self, key, listener: Callable):
        self.__listeners[key[0]][listener] = partial(key[1], listener)

    def _remove_listener(self, key, listener: Callable):
        self.__listeners[key[0]].pop(listener)

    def __api_read(self, char, data):
        self.__decoder.add(data)

    def __new_packet(self, packet):
        # print('response ' + ' '.join([hex(c) for c in packet.build()]))
        key = packet.id
        queue = self.__waiting[key]
        while not queue.empty():
            queue.get().set_result(packet)
        for f in self.__listeners[key].values():
            threading.Thread(target=f, args=(packet,)).start()

    @classmethod
    def implements(cls, method, with_target=False):
        m = getattr(cls, method.__name__, None)
        if m is method:
            return with_target == cls._require_target
        if hasattr(m, '_partialmethod'):
            f = m._partialmethod
            return f.func is method and (
                    ('proc' in f.keywords and not with_target) or with_target == cls._require_target)
        return False


class ToyV2(Toy):
    _packet = PacketV2
    _handshake = []

    _response_uuid = _send_uuid = '00010002-574f-4f20-5370-6865726f2121' #Original
    #_response_uuid = '22bb746f-2ba6-7554-2d6f-726568705327'
    #_send_uuid =     '22bb746f-2ba1-7554-2d6f-726568705327'
    #_send_uuid =     '00010002-574f-4f20-5370-6865726f2121'
    #_response_uuid = '00020002-574f-4f20-5370-6865726f2121'

    # _response_uuid = _send_uuid = '00010001-574f-4f20-5370-6865726f2121' #Available responses
    # _response_uuid = _send_uuid = '00010002-574f-4f20-5370-6865726f2121'
    # _response_uuid = _send_uuid = '00010003-574f-4f20-5370-6865726f2121'
    # _response_uuid = _send_uuid = '00020001-574f-4f20-5370-6865726f2121'
    # _response_uuid = _send_uuid = '00020002-574f-4f20-5370-6865726f2121'
    # _response_uuid = _send_uuid = '00020004-574f-4f20-5370-6865726f2121'
    # _response_uuid = _send_uuid = '00020005-574f-4f20-5370-6865726f2121'
    # _response_uuid = _send_uuid = '22bb746f-2bbd-7554-2d6f-726568705327'
    # _response_uuid = _send_uuid = '22bb746f-2bb2-7554-2d6f-726568705327'
    # _response_uuid = _send_uuid = '22bb746f-2bbf-7554-2d6f-726568705327'
    # _response_uuid = _send_uuid = '22bb746f-2ba0-7554-2d6f-726568705327'
    # _response_uuid = _send_uuid = '22bb746f-2ba1-7554-2d6f-726568705327'
    # _response_uuid = _send_uuid = '22bb746f-2ba6-7554-2d6f-726568705327'
    # _response_uuid = _send_uuid = '22bb746f-2bb0-7554-2d6f-726568705327'

#Some values found in apk - should revisit what values may be per robot
#Adaptor.BLEService = "22bb746f2bb075542d6f726568705327";
#Adaptor.WakeCharacteristic = "22bb746f2bbf75542d6f726568705327";
#Adaptor.TXPowerCharacteristic = "22bb746f2bb275542d6f726568705327";
#Adaptor.AntiDosCharacteristic = "22bb746f2bbd75542d6f726568705327";
#Adaptor.RobotControlService = "22bb746f2ba075542d6f726568705327";
#Adaptor.CommandsCharacteristic = "22bb746f2ba175542d6f726568705327";
#Adaptor.ResponseCharacteristic = "22bb746f2ba675542d6f726568705327";


from typing import NamedTuple


class CollisionDetected(NamedTuple):
    acceleration_x: float
    acceleration_y: float
    acceleration_z: float
    x_axis: bool
    y_axis: bool
    power_x: int
    power_y: int
    speed: int
    time: float



import struct
from enum import IntFlag, IntEnum

from spherov2.commands import Commands
from spherov2.helper import to_bytes
from spherov2.listeners.drive import MotorStall


class DriveFlags(IntFlag):
    FORWARD = 0x0  # 0b0
    BACKWARD = 0x1  # 0b1
    TURBO = 0x2  # 0b10
    FAST_TURN = 0x4  # 0b100
    LEFT_DIRECTION = 0x8  # 0b1000
    RIGHT_DIRECTION = 0x10  # 0b10000
    ENABLE_DRIFT = 0x20  # 0b100000


class RCDriveFlags(IntFlag):
    SLEW_LINEAR_VELOCITY = 0x1  # 0b1


class XYPositionDriveFlags(IntFlag):
    FORCE_REVERSE = 0x1  # 0b1
    AUTO_REVERSE = 0x2  # 0b10


class StabilizationIndexes(IntEnum):
    NO_CONTROL_SYSTEM = 0
    FULL_CONTROL_SYSTEM = 1
    PITCH_CONTROL_SYSTEM = 2
    ROLL_CONTROL_SYSTEM = 3
    YAW_CONTROL_SYSTEM = 4
    SPEED_AND_YAW_CONTROL_SYSTEM = 5


class RawMotorModes(IntEnum):
    OFF = 0
    FORWARD = 1
    REVERSE = 2


class GenericRawMotorIndexes(IntEnum):
    LEFT_DRIVE = 0
    RIGHT_DRIVE = 1
    HEAD = 2
    LEG = 3


class GenericRawMotorModes(IntEnum):
    MOTOR_OFF = 0
    FORWARD = 1
    REVERSE = 2


class LinearVelocitySlewMethods(IntEnum):
    CONSTANT = 0
    PROPORTIONAL = 1


class MotorIndexes(IntEnum):
    LEFT_MOTOR_INDEX = 0
    RIGHT_MOTOR_INDEX = 1


class Drive(Commands):
    _did = 22

    @staticmethod
    def set_raw_motors(toy, left_mode: RawMotorModes, left_speed, right_mode: RawMotorModes, right_speed, proc=None):
        toy._execute(Drive._encode(toy, 1, proc, [left_mode, left_speed, right_mode, right_speed]))

    @staticmethod
    def reset_yaw(toy, proc=None):
        toy._execute(Drive._encode(toy, 6, proc))

    @staticmethod
    def drive_with_heading(toy, speed, heading, drive_flags: DriveFlags, proc=None):
        toy._execute(Drive._encode(toy, 7, proc, [speed, *to_bytes(heading, 2), drive_flags]))

    @staticmethod
    def generic_raw_motor(toy, index: GenericRawMotorIndexes, mode: GenericRawMotorModes, speed, proc=None):
        toy._execute(Drive._encode(toy, 11, proc, [index, mode, *speed]))

    @staticmethod
    def set_stabilization(toy, stabilization_index: StabilizationIndexes, proc=None):
        toy._execute(Drive._encode(toy, 12, proc, [stabilization_index]))

    @staticmethod
    def set_control_system_type(toy, s, s2, proc=None):  # unknown name
        toy._execute(Drive._encode(toy, 14, proc, [s, s2]))

    @staticmethod
    def set_pitch_torque_modification_value(toy, f, proc=None):  # Untested / Unknown Param Name
        toy._execute(Drive._encode(toy, 15, proc, [f]))

    @staticmethod
    def set_component_parameters(toy, s, s2, f_arr, proc=None):  # unknown name
        toy._execute(Drive._encode(toy, 32, proc, [s, s2, *struct.pack('>%df' % len(f_arr), *f_arr)]))

    @staticmethod
    def get_component_parameters(toy, s, s2, proc=None):  # unknown name
        data = toy._execute(Drive._encode(toy, 33, proc, [s, s2])).data
        return struct.unpack('>%df' % (len(data) // 4), data)

    @staticmethod
    def set_custom_control_system_timeout(toy, timeout, proc=None):
        toy._execute(Drive._encode(toy, 34, proc, to_bytes(timeout, 2)))

    @staticmethod
    def enable_motor_stall_notify(toy, enable, proc=None):
        toy._execute(Drive._encode(toy, 37, proc, [int(enable)]))

    motor_stall_notify = (22, 38, 0xff), lambda listener, p: listener(MotorStall(*p.data))

    @staticmethod
    def enable_motor_fault_notify(toy, enable, proc=None):
        toy._execute(Drive._encode(toy, 39, proc, [int(enable)]))

    motor_fault_notify = (22, 40, 0xff), lambda listener, p: listener(bool(p.data[0]))

    @staticmethod
    def get_motor_fault_state(toy, proc=None):
        return bool(toy._execute(Drive._encode(toy, 41, proc)).data[0])



from collections import OrderedDict
from enum import IntEnum
from functools import partialmethod, lru_cache

from spherov2.commands.api_and_shell import ApiAndShell
from spherov2.commands.connection import Connection
from spherov2.commands.drive import Drive
from spherov2.commands.factory_test import FactoryTest
from spherov2.commands.firmware import Firmware
from spherov2.commands.io import IO
from spherov2.commands.power import Power
from spherov2.commands.sensor import Sensor
from spherov2.commands.system_info import SystemInfo
from spherov2.commands.system_mode import SystemMode
from spherov2.controls.v2 import DriveControl, FirmwareUpdateControl, LedControl, SensorControl, \
    StatsControl
from spherov2.toy import ToyV2, Toy, ToySensor
from spherov2.types import ToyType


class Mini(ToyV2):
    toy_type = ToyType('Sphero Mini', 'SM-', 'SM', .12)
    _handshake = [('00020005-574f-4f20-5370-6865726f2121', bytearray(b'usetheforce...band'))]  # Remove ForceBand

    class LEDs(IntEnum):
        AIMING = 0
        BODY_RED = 1
        BODY_GREEN = 2
        BODY_BLUE = 3
        USER_BODY_RED = 4
        USER_BODY_GREEN = 5
        USER_BODY_BLUE = 6

    # Sensors Available:
    sensors = OrderedDict(
        quaternion=OrderedDict(
            x=ToySensor(0x2000000, -1., 1.),
            y=ToySensor(0x1000000, -1., 1.),
            z=ToySensor(0x800000, -1., 1.),
            w=ToySensor(0x400000, -1., 1.)
        ),
        attitude=OrderedDict(
            pitch=ToySensor(0x40000, -179., 180.),
            roll=ToySensor(0x20000, -179., 180.),
            yaw=ToySensor(0x10000, -179., 180.)
        ),
        accelerometer=OrderedDict(
            x=ToySensor(0x8000, -8.19, 8.19),
            y=ToySensor(0x4000, -8.19, 8.19),
            z=ToySensor(0x2000, -8.19, 8.19)
        ),
        accel_one=OrderedDict(accel_one=ToySensor(0x200, 0., 8000.)),
        locator=OrderedDict(
            x=ToySensor(0x40, -32768., 32767., lambda x: x * 100.),
            y=ToySensor(0x20, -32768., 32767., lambda x: x * 100.),
        ),
        velocity=OrderedDict(
            x=ToySensor(0x10, -32768., 32767., lambda x: x * 100.),
            y=ToySensor(0x8, -32768., 32767., lambda x: x * 100.),
        ),
        speed=OrderedDict(speed=ToySensor(0x4, 0., 32767.)),
        core_time=OrderedDict(core_time=ToySensor(0x2, 0., 0.))
    )

    extended_sensors = OrderedDict(
        gyroscope=OrderedDict(
            x=ToySensor(0x2000000, -20000., 20000.),
            y=ToySensor(0x1000000, -20000., 20000.),
            z=ToySensor(0x800000, -20000., 20000.)
        )
    )

    # Mini Supported calls
    # API and Shell:
    ping = ApiAndShell.ping  # PingCommand
    get_api_protocol_version = ApiAndShell.get_api_protocol_version  # GetApiProtocolVersionCommand
    send_command_to_shell = ApiAndShell.send_command_to_shell  # SendCommandToShellCommand
    add_send_string_to_console_listener = partialmethod(Toy._add_listener,
                                                        ApiAndShell.send_string_to_console)  # SendStringToConsoleCommand
    remove_send_string_to_console_listener = partialmethod(Toy._remove_listener,
                                                           ApiAndShell.send_string_to_console)  # SendStringToConsoleCommand

    # Connection
    get_bluetooth_name = Connection.get_bluetooth_name  # GetBluetoothNameCommand
    set_bluetooth_name = Connection.set_bluetooth_name  # SetBluetoothNameCommand

    # Drive
    set_raw_motors = Drive.set_raw_motors  # SetRawMotorsCommand
    reset_yaw = Drive.reset_yaw  # ResetYawCommand
    drive_with_heading = Drive.drive_with_heading  # DriveWithHeadingCommand
    set_stabilization = Drive.set_stabilization  # SetStabilizationCommand

    # FactoryTest
    exit_factory_mode = FactoryTest.exit_factory_mode  # ExitFactoryModeCommand
    get_chassis_id = FactoryTest.get_chassis_id  # GetChassisIdCommand

    # Firmware
    get_pending_update_flags = Firmware.get_pending_update_flags  # GetPendingUpdateFlagsCommand

    # IO
    set_all_leds_with_16_bit_mask = IO.set_all_leds_with_16_bit_mask  # SetAllLedsWith16BitMaskCommand
    start_idle_led_animation = IO.start_idle_led_animation  # StartIdleLedAnimationCommand

    # Power
    add_battery_state_changed_notify_listener = partialmethod(Toy._add_listener,
                                                              Power.battery_state_changed_notify)  # BatteryStateChangedNotifyCommand
    remove_battery_state_changed_notify_listener = partialmethod(Toy._remove_listener,
                                                                 Power.battery_state_changed_notify)  # BatteryStateChangedNotifyCommand
    add_did_sleep_notify_listener = partialmethod(Toy._add_listener,
                                                  Power.did_sleep_notify)  # DidSleepNotifyCommand
    remove_did_sleep_notify_listener = partialmethod(Toy._remove_listener,
                                                     Power.did_sleep_notify)  # DidSleepNotifyCommand
    add_will_sleep_notify_listener = partialmethod(Toy._add_listener,
                                                   Power.will_sleep_notify)  # WillSleepNotifyCommand
    remove_will_sleep_notify_listener = partialmethod(Toy._remove_listener,
                                                      Power.will_sleep_notify)  # WillSleepNotifyCommand
    enable_battery_state_changed_notify = Power.enable_battery_state_changed_notify  # EnableBatteryStateChangedNotifyCommand
    enter_deep_sleep = Power.enter_deep_sleep  # EnterDeepSleepCommand
    get_battery_percentage = Power.get_battery_percentage  # GetBatteryPercentageCommand
    get_battery_state = Power.get_battery_state  # GetBatteryStateCommand
    get_battery_voltage = Power.get_battery_voltage  # GetBatteryVoltageCommand
    get_battery_voltage_state = Power.get_battery_voltage_state  # GetBatterVoltageCommand
    sleep = Power.sleep  # SleepCommand
    wake = Power.wake  # WakeCommand

    # Sensor
    add_collision_detected_notify_listener = partialmethod(Toy._add_listener,
                                                           Sensor.collision_detected_notify)  # CollisionDetectedNotifyCommand
    remove_collision_detected_notify_listener = partialmethod(Toy._remove_listener,
                                                              Sensor.collision_detected_notify)  # CollisionDetectedNotifyCommand
    configure_collision_detection = Sensor.configure_collision_detection  # ConfigureCollisionDetectionCommand
    enable_collision_detected_notify = Sensor.enable_collision_detected_notify  # EnableCollisionDetectedNotifyCommand
    enable_gyro_max_notify = Sensor.enable_gyro_max_notify  # EnableGyroMaxNotifyCommand
    get_extended_sensor_streaming_mask = Sensor.get_extended_sensor_streaming_mask  # GetExtendedSensorStreamingMaskCommand
    set_extended_sensor_streaming_mask = Sensor.set_extended_sensor_streaming_mask  # SetExtendedSensorStreamingMaskCommand
    get_sensor_streaming_mask = Sensor.get_sensor_streaming_mask  # SetSensorStreamingMaskCommand
    set_sensor_streaming_mask = Sensor.set_sensor_streaming_mask  # GetSensorStreamingMaskCommand
    add_gyro_max_notify_listener = partialmethod(Toy._add_listener, Sensor.gyro_max_notify)  # GyroMaxNotifyCommand
    remove_gyro_max_notify_listener = partialmethod(Toy._remove_listener,
                                                    Sensor.gyro_max_notify)  # GyroMaxNotifyCommand
    reset_locator_x_and_y = Sensor.reset_locator_x_and_y  # ResetLocationXAndYCommand
    add_sensor_streaming_data_notify_listener = partialmethod(Toy._add_listener,
                                                              Sensor.sensor_streaming_data_notify)  # SensorStreamingDataNotifyCommand
    remove_sensor_streaming_data_notify_listener = partialmethod(Toy._remove_listener,
                                                                 Sensor.sensor_streaming_data_notify)  # SensorStreamingDataNotifyCommand
    set_locator_flags = Sensor.set_locator_flags  # SetLocatorFlagsCommand

    # System Info
    get_board_revision = SystemInfo.get_board_revision  # GetBoardRevisionCommand
    get_bootloader_version = SystemInfo.get_bootloader_version  # GetBootloaderVersionCommand
    get_mac_address = SystemInfo.get_mac_address  # GetMacAddressCommand
    get_main_app_version = SystemInfo.get_main_app_version  # GetMainAppVersionCommand
    get_model_number = SystemInfo.get_model_number  # GetModelNumberCommand
    get_processor_name = SystemInfo.get_processor_name  # GetProcessorNameCommand
    get_stats_id = SystemInfo.get_stats_id  # GetStatsIdCommand
    get_three_character_sku = SystemInfo.get_three_character_sku  # GetThreeCharacterSkuCommand

    # System Mode
    enable_desktoy_mode = SystemMode.enable_desktoy_mode  # EnableDesktoyModeCommand

    # Controls - V2
    @property
    @lru_cache(None)
    def drive_control(self):
        return DriveControl(self)

    @property
    @lru_cache(None)
    def multi_led_control(self):
        return LedControl(self)

    @property
    @lru_cache(None)
    def sensor_control(self):
        return SensorControl(self)

    @property
    @lru_cache(None)
    def stats_control(self):
        return StatsControl(self)

    @property
    @lru_cache(None)
    def firmware_update_control(self):
        return FirmwareUpdateControl(self)



from typing import NamedTuple


class Options(NamedTuple):
    disable_sleep_in_charger: bool
    enable_vector_drive: bool
    disable_self_level_in_charger: bool
    enable_tail_light_always_on: bool
    enable_motion_timeout: bool
    enable_gyro_max_notify: bool
    enable_full_speed: bool



import math
import threading
import time
from collections import namedtuple, defaultdict
from enum import Enum, IntEnum, auto
from functools import partial
from typing import Union, Callable, Dict, Iterable

import numpy as np
from transforms3d.euler import euler2mat

from spherov2.commands.animatronic import R2LegActions
from spherov2.commands.io import IO
from spherov2.commands.power import BatteryVoltageAndStateStates
from spherov2.controls import RawMotorModes
from spherov2.helper import bound_value, bound_color
from spherov2.toy import Toy
from spherov2.toy.bb8 import BB8
from spherov2.toy.bb9e import BB9E
from spherov2.toy.bolt import BOLT
from spherov2.toy.mini import Mini
from spherov2.toy.ollie import Ollie
from spherov2.toy.r2d2 import R2D2
from spherov2.toy.r2q5 import R2Q5
from spherov2.toy.rvr import RVR
from spherov2.toy.sphero import Sphero
from spherov2.types import Color
from spherov2.utils import ToyUtil


class Stance(str, Enum):
    Bipod = 'twolegs'
    Tripod = 'threelegs'


class EventType(Enum):
    on_collision = auto()  # [f.Sphero, f.Ollie, f.BB8, f.BB9E, f.R2D2, f.R2Q5, f.BOLT, f.Mini]
    on_freefall = auto()  # [f.Sphero, f.Ollie, f.BB8, f.BB9E, f.R2D2, f.R2Q5, f.BOLT, f.Mini]
    on_landing = auto()  # [f.Sphero, f.Ollie, f.BB8, f.BB9E, f.R2D2, f.R2Q5, f.BOLT, f.Mini]
    on_gyro_max = auto()  # [f.Sphero, f.Mini, f.Ollie, f.BB8, f.BB9E, f.BOLT, f.Mini]
    on_charging = auto()  # [f.Sphero, f.Ollie, f.BB8, f.BB9E, f.R2D2, f.R2Q5, f.BOLT]
    on_not_charging = auto()  # [f.Sphero, f.Ollie, f.BB8, f.BB9E, f.R2D2, f.R2Q5, f.BOLT]
    on_magnetometer_north_yaw = auto()  # [f.BOLT] TODO
    on_sensor_streaming_data = auto()  # [f.BOLT] TODO
    on_ir_message = auto()  # [f.BOLT, f.RVR] TODO
    on_color = auto()  # [f.RVR] TODO


class LedManager:
    def __init__(self, cls):
        if cls is RVR:
            self.__mapping = {
                'front': ('left_headlight', 'right_headlight'),
                'main': ('left', 'right', 'front', 'back')
            }
        elif cls in (R2D2, R2Q5, BOLT):
            self.__mapping = {'main': ('front', 'back')}
        else:
            self.__mapping = {}
        self.__leds = defaultdict(partial(Color, 0, 0, 0))

    def __setitem__(self, key, value):
        if key in self.__mapping:
            for led in self.__mapping[key]:
                self.__setitem__(led, value)
        else:
            self.__leds[key] = value

    def __getitem__(self, item):
        if item in self.__mapping:
            return self.__getitem__(self.__mapping[item][0])
        return self.__leds[item]

    def get(self, item, default):
        if item in self.__mapping:
            return self.get(self.__mapping[item][0], default)
        return self.__leds.get(item, default)


rawMotor = namedtuple('rawMotor', ('left', 'right'))


class SpheroEduAPI:
    """Implementation of Sphero Edu Javascript APIs: https://sphero.docsapp.io/docs/get-started"""

    def __init__(self, toy: Toy):
        self.__toy = toy
        self.__heading = 0
        self.__speed = 0
        self.__stabilization = True
        self.__raw_motor = rawMotor(0, 0)
        self.__leds = LedManager(toy.__class__)

        self.__sensor_data: Dict[str, Union[float, Dict[str, float]]] = {'distance': 0., 'color_index': -1}
        self.__sensor_name_mapping = {}
        self.__last_location = (0., 0.)
        self.__last_non_fall = time.time()
        self.__falling_v = 1.
        self.__last_message = None
        self.__should_land = self.__free_falling = False

        self.__listeners = defaultdict(set)
        ToyUtil.add_listeners(toy, self)

        self.__stopped = threading.Event()
        self.__stopped.set()
        self.__updating = threading.Lock()
        self.__thread = None

    def __enter__(self):
        self.__stopped.clear()
        self.__thread = threading.Thread(target=self.__background)
        self.__toy.__enter__()
        self.__thread.start()
        try:
            self.__toy.wake()
            ToyUtil.set_robot_state_on_start(self.__toy)
            self.__start_capturing_sensor_data()
        except:
            self.__exit__(None, None, None)
            raise
        return self

    def __exit__(self, *args):
        self.__stopped.set()
        self.__thread.join()
        try:
            ToyUtil.sleep(self.__toy)
        except:
            pass
        self.__toy.__exit__(*args)

    def __background(self):
        while not self.__stopped.wait(0.8):
            with self.__updating:
                self.__update_speeds()

    def _will_sleep_notify(self):
        ToyUtil.ping(self.__toy)

    # Movements: control the robot's motors and control system.
    def __update_speeds(self):
        if self.__speed != 0:
            self.__update_speed()
        if self.__raw_motor.left != 0 or self.__raw_motor.right != 0:
            self.__update_raw_motor()

    def __stop_all(self):
        if self.__speed != 0:
            self.__speed = 0
            self.__update_speed()
        if self.__raw_motor.left != 0 or self.__raw_motor.right != 0:
            self.__raw_motor = rawMotor(0, 0)
            self.__update_raw_motor()

    def roll(self, heading: int, speed: int, duration: float):
        """Combines heading(0-360°), speed(-255-255), and duration to make the robot roll with one line of code.
        For example, to have the robot roll at 90°, at speed 200 for 2s, use ``roll(90, 200, 2)``"""
        if isinstance(self.__toy, Mini) and speed != 0:
            speed = round((speed + 126) * 2 / 3) if speed > 0 else round((speed - 126) * 2 / 3)
        self.__speed = bound_value(-255, speed, 255)
        self.__heading = heading % 360
        if speed < 0:
            self.__heading = (self.__heading + 180) % 360
        self.__update_speed()
        time.sleep(duration)
        self.stop_roll()

    def __update_speed(self):
        ToyUtil.roll_start(self.__toy, self.__heading, self.__speed)

    def set_speed(self, speed: int):
        """Sets the speed of the robot from -255 to 255, where positive speed is forward, negative speed is backward,
        and 0 is stopped. Each robot type translates this value differently into a real world speed;
        Ollie is almost three times faster than Sphero. For example, use ``set_speed(188)`` to set the speed to 188
        which persists until you set a different speed. You can also read the real-time velocity value in centimeters
        per second reported by the motor encoders.
        """
        if isinstance(self.__toy, Mini) and speed != 0:
            speed = round((speed + 126) * 2 / 3) if speed > 0 else round((speed - 126) * 2 / 3)
        self.__speed = bound_value(-255, speed, 255)
        self.__update_speed()

    def stop_roll(self, heading: int = None):
        """Sets the speed to zero to stop the robot, effectively the same as the ``set_speed(0)`` command."""
        if heading is not None:
            self.__heading = heading % 360
        self.__speed = 0
        ToyUtil.roll_stop(self.__toy, self.__heading, False)

    def set_heading(self, heading: int):
        """Sets the direction the robot rolls.
        Assuming you aim the robot with the blue tail light facing you, then 0° is forward, 90° is right,
        270° is left, and 180° is backward. For example, use ``set_heading(90)`` to face right."""
        self.__heading = heading % 360
        ToyUtil.roll_start(self.__toy, self.__heading, self.__speed)

    def spin(self, angle: int, duration: float):
        """Spins the robot for a given number of degrees over time, with 360° being a single revolution.
        For example, to spin the robot 360° over 1s, use: ``spin(360, 1)``.
        Use :func:`set_speed` prior to :func:`spin` to have the robot move in circle or an arc or circle.

        Note: Unlike official API, performance of spin is guaranteed, but may be longer than the specified duration."""

        if angle == 0:
            return

        time_pre_rev = .45

        if isinstance(self.__toy, RVR):
            time_pre_rev = 1.5
        elif isinstance(self.__toy, (R2D2, R2Q5)):
            time_pre_rev = .7
        elif isinstance(self.__toy, Mini):
            time_pre_rev = .5
        elif isinstance(self.__toy, Ollie):
            time_pre_rev = .6

        abs_angle = abs(angle)
        duration = max(duration, time_pre_rev * abs_angle / 360)

        start = time.time()
        angle_gone = 0
        with self.__updating:
            while angle_gone < abs_angle:
                delta = round(min((time.time() - start) / duration, 1.) * abs_angle) - angle_gone
                self.set_heading(self.__heading + delta if angle > 0 else self.__heading - delta)
                angle_gone += delta

    def set_stabilization(self, stabilize: bool):
        """Turns the stabilization system on and ``set_stabilization(false)`` turns it off.
        Stabilization is normally on to keep the robot upright using the Inertial Measurement Unit (IMU),
        a combination of readings from the Accelerometer (directional acceleration), Gyroscope (rotation speed),
        and Encoders (location and distance). When ``set_stabilization(false)`` and you power the motors,
        the robot will not balance, resulting in possible unstable behaviors like wobbly driving,
        or even jumping if you set the power very high. Some use cases to turn it off are:

        1. Jumping: Set Motor Power to max values and the robot will jump off the ground!
        2. Gyro: Programs like the Spinning Top where you want to to isolate the Gyroscope readings rather than having
           the robot auto balance inside the shell.

        When stabilization is off you can't use :func:`set_speed` to set a speed because it requires the control system
        to be on to function. However, you can control the motors using Motor Power with :func:`raw_motor` when
        the control system is off."""
        self.__stabilization = stabilize
        if isinstance(self.__toy, (Sphero, Mini, Ollie, BB8, BB9E, BOLT)):
            ToyUtil.set_stabilization(self.__toy, stabilize)

    def __update_raw_motor(self):
        ToyUtil.set_raw_motor(self.__toy,
                              RawMotorModes.REVERSE if self.__raw_motor.left < 0 else RawMotorModes.FORWARD,
                              abs(self.__raw_motor.left),
                              RawMotorModes.REVERSE if self.__raw_motor.right < 0 else RawMotorModes.FORWARD,
                              abs(self.__raw_motor.right))

    def raw_motor(self, left: int, right: int, duration: float):
        """Controls the electrical power sent to the left and right motors independently, on a scale from -255 to 255
        where positive is forward, negative is backward, and 0 is stopped. If you set both motors to full power
        the robot will jump because stabilization (use of the IMU to keep the robot upright) is disabled when using
        this command. This is different from :func:`set_speed` because Raw Motor sends an "Electromotive force"
        to the motors, whereas Set Speed is a target speed measured by the encoders. For example, to set the raw motor
        to full power for 4s, making the robot jump off the ground, use ``raw_motor(255, 255, 4)``."""
        stabilize = self.__stabilization
        if stabilize:
            self.set_stabilization(False)
        self.__raw_motor = rawMotor(bound_value(-255, left, 255), bound_value(-255, right, 255))
        self.__update_raw_motor()
        if duration is not None:
            time.sleep(duration)
            if stabilize:
                self.set_stabilization(True)
            self.__raw_motor = rawMotor(0, 0)
            ToyUtil.set_raw_motor(self.__toy, RawMotorModes.OFF, 0, RawMotorModes.OFF, 0)

    def reset_aim(self):
        """Resets the heading calibration (aim) angle to use the current direction of the robot as 0°."""
        ToyUtil.reset_heading(self.__toy)

    # Star Wars Droid Movements
    def play_animation(self, animation: IntEnum):
        """Plays iconic `Star Wars Droid animations <https://edu.sphero.com/remixes/1195472/>`_ unique to BB-8, BB-9E,
        R2-D2 and R2-Q5 that combine movement, lights and sound. All animation enums can be accessed under the droid
        class, such as :class:`R2D2.Animations.CHARGER_1`."""
        if hasattr(self.__toy, 'Animations'):
            if animation not in self.__toy.Animations:
                raise ValueError(f'Animation {animation} cannot be played by this toy')
            with self.__updating:
                self.__stop_all()
            ToyUtil.play_animation(self.__toy, animation, True)

    # The R2-D2 and R2-Q5 Droids are physically different from other Sphero robots,
    # so there are some unique commands that only they can use.
    def set_dome_position(self, angle: float):
        """Rotates the dome on its axis, from -160° to 180°. For example, set to 45° using ``set_dome_position(45).``"""
        if isinstance(self.__toy, (R2D2, R2Q5)):
            ToyUtil.set_head_position(self.__toy, bound_value(-160., angle, 180.))

    def set_stance(self, stance: Stance):
        """Changes the stance between bipod and tripod. Set to bipod using ``set_stance(Stance.Bipod)`` and
        to tripod using ``set_stance(Stance.Tripod)``. Tripod is required for rolling."""
        if isinstance(self.__toy, (R2D2, R2Q5)):
            if stance == Stance.Bipod:
                ToyUtil.perform_leg_action(self.__toy, R2LegActions.TWO_LEGS)
            elif stance == Stance.Tripod:
                ToyUtil.perform_leg_action(self.__toy, R2LegActions.THREE_LEGS)
            else:
                raise ValueError(f'Stance {stance} is not supported')

    def set_waddle(self, waddle: bool):
        """Turns the waddle walk on using `set_waddle(True)`` and off using ``set_waddle(False)``."""
        if isinstance(self.__toy, (R2D2, R2Q5)):
            with self.__updating:
                self.__stop_all()
            ToyUtil.perform_leg_action(self.__toy, R2LegActions.WADDLE if waddle else R2LegActions.STOP)

    # Lights: control the color and brightness of LEDs on a robot.
    def set_main_led(self, color: Color):
        """Changes the color of the main LED light, or the full matrix on Sphero BOLT. Set this using RGB
        (red, green, blue) values on a scale of 0 - 255. For example, ``set_main_led(Color(r=90, g=255, b=90))``."""
        self.__leds['main'] = bound_color(color, self.__leds['main'])
        ToyUtil.set_main_led(self.__toy, **self.__leds['main']._asdict(), is_user_color=False)

    def set_front_led(self, color: Color):
        """For Sphero RVR: Changes the color of RVR's front two LED headlights together.

        For Sphero BOLT, R2D2, R2Q5: Changes the color of the front LED light.

        Set this using RGB (red, green, blue) values on a scale of 0 - 255. For example, the magenta color is expressed
        as ``set_front_color(Color(239, 0, 255))``."""
        if isinstance(self.__toy, (R2D2, R2Q5, BOLT, RVR)):
            self.__leds['front'] = bound_color(color, self.__leds['front'])
            ToyUtil.set_front_led(self.__toy, **self.__leds['front']._asdict())

    def set_back_led(self, color: Union[Color, int]):
        """For older Sphero:
        Sets the brightness of the back aiming LED, aka the "Tail Light." This LED is limited to blue only, with a
        brightness scale from 0 to 255. For example, use ``set_back_led(255)`` to set the back LED to full brightness.
        Use :func:`time.sleep` to set it on for a duration. For example, to create a dim and a bright blink
        sequence use::

            set_back_led(0)  # Dim
            delay(0.33)
            set_back_led(255)  # Bright
            delay(0.33)

        For Sphero BOLT, R2D2, R2Q5:
        Changes the color of the back LED light. Set this using RGB (red, green, blue) values on a scale of 0 - 255.

        For Sphero RVR:
        Changes the color of the left and right breaklight LED light. Set this using RGB (red, green, blue) values
        on a scale of 0 - 255."""
        if isinstance(color, int):
            self.__leds['back'] = Color(0, 0, bound_value(0, color, 255))
            ToyUtil.set_back_led_brightness(self.__toy, self.__leds['back'].b)
        elif isinstance(self.__toy, (R2D2, R2Q5, BOLT, RVR, Mini)):
            self.__leds['back'] = bound_color(color, self.__leds['back'])
            ToyUtil.set_back_led(self.__toy, **self.__leds['back']._asdict())

    def fade(self, from_color: Color, to_color: Color, duration: float):
        """Changes the main LED lights from one color to another over a period of seconds. For example, to fade from
        green to blue over 3s, use: ``fade(Color(0, 255, 0), Color(0, 0, 255), 3.0)``."""
        from_color = bound_color(from_color, self.__leds['main'])
        to_color = bound_color(to_color, self.__leds['main'])

        start = time.time()
        while True:
            frac = (time.time() - start) / duration
            if frac >= 1:
                break
            self.set_main_led(Color(
                r=round(from_color.r * (1 - frac) + to_color.r * frac),
                g=round(from_color.g * (1 - frac) + to_color.g * frac),
                b=round(from_color.b * (1 - frac) + to_color.b * frac)))
        self.set_main_led(to_color)

    def strobe(self, color: Color, period: float, count: int):
        """Repeatedly blinks the main LED lights. The period is the time, in seconds, the light stays on during a
        single blink; cycles is the total number of blinks. The time for a single cycle is twice the period
        (time for a blink plus the same amount of time for the light to be off). Another way to say this is the period
        is 1/2 the time it takes for a single cycle. So, to strobe red 15 times in 3 seconds, use:
        ``strobe(Color(255, 57, 66), (3 / 15) * .5, 15)``."""
        for i in range(count * 2):
            if i & 1:
                self.set_main_led(color)
            else:
                self.set_main_led(Color(0, 0, 0))
            time.sleep(period)

    # TODO Sphero BOLT Lights
    def set_matrix_pixel(self, x: int, y: int, color: Color):
        """For Sphero BOLT: Changes the color of BOLT's matrix at X and Y value. 8x8
        """
        strMapLoc: str = str(x) + ':' + str(y)
        self.__leds[strMapLoc] = bound_color(color, self.__leds[
            strMapLoc])  # TODO: Do this in a way that works with line and fill
        ToyUtil.set_matrix_pixel(self.__toy, x, y, **self.__leds[strMapLoc]._asdict(), is_user_color=False)

    def set_matrix_line(self, x1: int, y1: int, x2: int, y2: int, color: Color):
        """For Sphero BOLT: Changes the color of BOLT's matrix from x1,y1 to x2,y2 in a line. 8x8
        """
        strMapLoc: str = str(x1) + 'x' + str(y1) + '|' + str(x2) + 'x' + str(y2)
        self.__leds[strMapLoc] = bound_color(color, self.__leds[
            strMapLoc])  # TODO: Do this in a way that works with pixel and fill (needs to be accurate to diagonal lines)
        ToyUtil.set_matrix_line(self.__toy, x1, y1, x2, y2, **self.__leds[strMapLoc]._asdict(), is_user_color=False)

    def set_matrix_fill(self, x1: int, y1: int, x2: int, y2: int, color: Color):
        """For Sphero BOLT: Changes the color of BOLT's matrix from x1,y1 to x2,y2 in a box. 8x8
        """
        strMapLoc: str = str(x1) + 'x' + str(y1) + '[]' + str(x2) + 'x' + str(y2)
        self.__leds[strMapLoc] = bound_color(color, self.__leds[
            strMapLoc])  # TODO: Do this in a way that works with pixel and line
        ToyUtil.set_matrix_fill(self.__toy, x1, y1, x2, y2, **self.__leds[strMapLoc]._asdict(), is_user_color=False)

    def register_matrix_animation(self, s, s2, z, s3, s_arr, i, i_arr):  # TODO: fix this function
        ToyUtil.register_matrix_animation(self.__toy, s, s2, z, s3, s_arr, i, i_arr)

    def play_matrix_animation(self, s):
        ToyUtil.play_matrix_animation(self.__toy, s)

    # Sphero RVR Lights
    def set_left_headlight_led(self, color: Color):
        """Changes the color of the front left headlight LED on RVR. Set this using RGB (red, green, blue) values on a
        scale of 0 - 255. For example, the pink color is expressed as
        ``set_left_headlight_led(Color(253, 159, 255))``."""
        if isinstance(self.__toy, RVR):
            self.__leds['left_headlight'] = bound_color(color, self.__leds['left_headlight'])
            ToyUtil.set_left_front_led(self.__toy, **self.__leds['left_headlight']._asdict())

    def set_right_headlight_led(self, color: Color):
        """Changes the color of the front right headlight LED on RVR. Set this using RGB (red, green, blue) values on a
        scale of 0 - 255. For example, the blue color is expressed as
        ``set_right_headlight_led(0, 28, 255)``."""
        if isinstance(self.__toy, RVR):
            self.__leds['right_headlight'] = bound_color(color, self.__leds['right_headlight'])
            ToyUtil.set_right_front_led(self.__toy, **self.__leds['right_headlight']._asdict())

    def set_left_led(self, color: Color):
        """Changes the color of the LED on RVR's left side (which is the side with RVR's battery bay door). Set this
        using RGB (red, green, blue) values on a scale of 0 - 255. For example, the green color is expressed as
        ``set_left_led(Color(0, 255, 34))``."""
        if isinstance(self.__toy, RVR):
            self.__leds['left'] = bound_color(color, self.__leds['left'])
            ToyUtil.set_battery_side_led(self.__toy, **self.__leds['left']._asdict())

    def set_right_led(self, color: Color):
        """Changes the color of the LED on RVR's right side (which is the side with RVR's power button). Set this using
        RGB (red, green, blue) values on a scale of 0 - 255. For example, the red color is expressed as
        ``set_right_led(Color(255, 18, 0))``."""
        if isinstance(self.__toy, RVR):
            self.__leds['right'] = bound_color(color, self.__leds['right'])
            ToyUtil.set_power_side_led(self.__toy, **self.__leds['right']._asdict())

    # BB-9E Lights
    def set_dome_leds(self, brightness: int):
        """Controls the brightness of the two single color LEDs (red and blue) in the dome, from 0 to 15. We don't use
        0-255 for this light because it has less granular control. For example, set them to full brightness using
        ``set_dome_leds(15)``."""
        if isinstance(self.__toy, BB9E):
            self.__leds['dome'] = bound_value(0, brightness, 15)
            ranged = self.__leds['dome'] * 255 // 15
            ToyUtil.set_head_led(self.__toy, ranged)

    # R2-D2 & R2-Q5 Lights
    def set_holo_projector_led(self, brightness: int):
        """Changes the brightness of the Holographic Projector white LED, from 0 to 255. For example, set it to full
        brightness using ``set_holo_projector_led(255)``."""
        if isinstance(self.__toy, (R2D2, R2Q5)):
            self.__leds['holo_projector'] = bound_value(0, brightness, 255)
            ToyUtil.set_holo_projector(self.__toy, self.__leds['holo_projector'])

    def set_logic_display_leds(self, brightness: int):
        """Changes the brightness of the Logic Display LEDs, from 0 to 255. For example, set it to full brightness
        using ``set_logic_display_leds(255)``."""
        if isinstance(self.__toy, (R2D2, R2Q5)):
            self.__leds['logic_display'] = bound_value(0, brightness, 255)
            ToyUtil.set_logic_display(self.__toy, self.__leds['logic_display'])

    # Sounds: Control sounds and words which can play from your programming device's speaker or the robot.
    def play_sound(self, sound: IntEnum):
        """Unique Star Wars Droid Sounds are available for BB-8, BB-9E and R2-D2. For example, to play the R2-D2 Burnout
        sound use ``play_sound(R2D2.Audio.R2_BURNOUT)``."""
        if hasattr(self.__toy, 'Audio'):
            if sound not in self.__toy.Audio:
                raise ValueError(f'Sound {sound} cannot be played by this toy')
            ToyUtil.play_sound(self.__toy, sound, False)

    # Sensors: Querying sensor data allows you to react to real-time values coming from the robots' physical sensors.
    def __start_capturing_sensor_data(self):
        if isinstance(self.__toy, RVR):
            sensors = ['accelerometer', 'gyroscope', 'imu', 'locator', 'velocity', 'ambient_light', 'color_detection']
            self.__sensor_name_mapping['imu'] = 'attitude'
        elif isinstance(self.__toy, BOLT):
            sensors = ['accelerometer', 'gyroscope', 'attitude', 'locator', 'velocity', 'ambient_light']
        else:
            sensors = ['attitude', 'accelerometer', 'gyroscope', 'locator', 'velocity']
        ToyUtil.enable_sensors(self.__toy, sensors)

    def _sensor_data_listener(self, sensor_data: Dict[str, Dict[str, float]]):
        for sensor, data in sensor_data.items():
            if sensor in self.__sensor_name_mapping:
                self.__sensor_data[self.__sensor_name_mapping[sensor]] = data
            else:
                self.__sensor_data[sensor] = data
        if 'attitude' in self.__sensor_data and 'accelerometer' in self.__sensor_data:
            att = self.__sensor_data['attitude']
            r = euler2mat(*np.deg2rad((att['roll'], att['pitch'], att['yaw'])), axes='szxy')
            acc = self.__sensor_data['accelerometer']
            self.__sensor_data['vertical_accel'] = -np.matmul(np.linalg.inv(r), (acc['x'], -acc['z'], acc['y']))[1]
            self.__process_falling(self.__sensor_data['vertical_accel'])
        if 'locator' in self.__sensor_data:
            cur_loc = self.__sensor_data['locator']
            cur_loc = (cur_loc['x'], cur_loc['y'])
            self.__sensor_data['distance'] += math.hypot(cur_loc[0] - self.__last_location[0],
                                                         cur_loc[1] - self.__last_location[1])
            self.__last_location = cur_loc
        if 'color_detection' in self.__sensor_data:
            color = self.__sensor_data['color_detection']
            index = color['index']
            if index != self.__sensor_data['color_index'] and index < 255 and color['confidence'] >= 0.71:
                self.__sensor_data['color_index'] = index
                self.__call_event_listener(EventType.on_color, Color(int(color['r']), int(color['g']), int(color['b'])))

    def __process_falling(self, a):
        self.__falling_v = (self.__falling_v + a * 3) / 4
        cur = time.time()
        if -.5 < self.__falling_v < .5 if self.__stabilization else -.1 < a < .1:
            if cur - self.__last_non_fall > .2 and not self.__free_falling:
                self.__call_event_listener(EventType.on_freefall)
                self.__free_falling = True
                self.__should_land = True
        else:
            self.__last_non_fall = cur
            self.__free_falling = False
        if self.__should_land and (
                self.__falling_v < -1.1 or self.__falling_v > 1.1 if self.__stabilization else a < -.8 or a > -.8):
            self.__call_event_listener(EventType.on_landing)
            self.__should_land = False

    def _collision_detected_notify(self, args):
        self.__call_event_listener(EventType.on_collision)

    def _battery_state_changed_notify(self, state: BatteryVoltageAndStateStates):
        if state == BatteryVoltageAndStateStates.CHARGED or state == BatteryVoltageAndStateStates.CHARGING:
            self.__call_event_listener(EventType.on_charging)
        else:
            self.__call_event_listener(EventType.on_not_charging)

    def _gyro_max_notify(self, flags):
        self.__call_event_listener(EventType.on_gyro_max)

    def _magnetometer_north_yaw_notify(self, flags):
        self.__call_event_listener(EventType.on_magnetometer_north_yaw)

    def _sensor_streaming_data_notify(self, flags):
        self.__call_event_listener(EventType.on_sensor_streaming_data)

    def get_acceleration(self):
        """Provides motion acceleration data along a given axis measured by the Accelerometer, in g's, where g =
        9.80665 m/s^2.

        ``get_acceleration()['x']`` is the left-to-right acceleration, from -8 to 8 g's.

        ``get_acceleration()['y']`` is the forward-to-back acceleration, from of -8 to 8 g's.

        ``get_acceleration()['z']`` is the upward-to-downward acceleration, from -8 to 8 g's."""

        return self.__sensor_data.get('accelerometer', None)

    def get_vertical_acceleration(self):
        """This is the upward or downward acceleration regardless of the robot's orientation, from -8 to 8 g's."""
        return self.__sensor_data.get('vertical_accel', None)

    def get_orientation(self):
        """Provides the tilt angle along a given axis measured by the Gyroscope, in degrees.

        ``get_orientation()['pitch']`` is the forward or backward tilt angle, from -180° to 180°.

        ``get_orientation()['roll']`` is left or right tilt angle, from -90° to 90°.

        ``get_orientation()['yaw']`` is the spin (twist) angle, from -180° to 180°."""
        return self.__sensor_data.get('attitude', None)

    def get_gyroscope(self):
        """Provides the rate of rotation around a given axis measured by the gyroscope, from -2,000° to 2,000°
        per second.

        ``get_gyroscope().['pitch']`` is the rate of forward or backward spin, from -2,000° to 2,000° per second.

        ``get_gyroscope().['roll']`` is the rate of left or right spin, from -2,000° to 2,000° per second.

        ``get_gyroscope().['yaw']`` is the rate of sideways spin, from -2,000° to 2,000° per second."""
        return self.__sensor_data.get('gyroscope', None)

    def get_velocity(self):
        """Provides the velocity along a given axis measured by the motor encoders, in centimeters per second.

        ``get_velocity()['x']`` is the right (+) or left (-) velocity, in centimeters per second.

        ``get_velocity()['y']`` is the forward (+) or back (-) velocity, in centimeters per second."""
        return self.__sensor_data.get('velocity', None)

    def get_location(self):
        """Provides the location where the robot is in space (x,y) relative to the origin, in centimeters. This is not
        the distance traveled during the program, it is the offset from the origin (program start).

        ``get_location()['x']`` is the right (+) or left (-) distance from the origin of the program start, in
        centimeters.

        ``get_location()['y']`` is the forward (+) or backward (-) distance from the origin of the program start, in
        centimeters."""
        return self.__sensor_data.get('locator', None)

    def get_distance(self):
        """Provides the total distance traveled in the program, in centimeters."""
        return self.__sensor_data.get('distance', None)

    def get_speed(self):
        """Provides the current target speed of the robot, from -255 to 255, where positive is forward, negative is
        backward, and 0 is stopped."""
        return self.__speed

    def get_heading(self):
        """Provides the target directional angle, in degrees. Assuming you aim the robot with the tail facing you,
        then 0° heading is forward, 90° is right, 180° is backward, and 270° is left."""
        return self.__heading

    def get_main_led(self):
        """Provides the RGB color of the main LEDs, from 0 to 255 for each color channel.

        ``get_main_led().r`` is the red channel, from 0 - 255.

        ``get_main_led().g`` is the green channel, from 0 - 255.

        ``get_main_led().b`` is the blue channel, from 0 - 255."""
        return self.__leds.get('main', None)

    # Sphero BOLT Sensors
    # TODO Compass Direction

    def get_luminosity_direct(self):
        """similar to get_luminosity, however this is a more direct call to the sphero to get a value directly"""
        return ToyUtil.get_ambient_light_sensor_value(self.__toy)

    def get_luminosity(self):
        """Provides the light intensity from 0 - 100,000 lux, where 0 lux is full darkness and 30,000-100,000 lux is
        direct sunlight. You may need to adjust a condition based on luminosity in different environments as light
        intensity can vary greatly between rooms."""
        return self.__sensor_data.get('ambient_light', None)

    def get_last_ir_message(self):
        """Returns which channel the last infrared message was received on. You need to declare the ``on_ir_message``
        event for each IR message you plan to see returned."""
        return self.__last_message

    def get_back_led(self):
        """Provides the RGB color of the back LED, from 0 to 255 for each color channel."""
        return self.__leds.get('back', None)

    def get_front_led(self):
        """Provides the RGB color of the front LED, from 0 to 255 for each color channel."""
        return self.__leds.get('front', None)

    # Sphero RVR Sensors
    def get_color(self):
        """Provides the RGB color, from 0 to 255 for each color channel, that is returned from RVR's color sensor.

        ``get_color().r`` is the red channel, from 0 - 255, that is returned from RVR's color sensor.

        ``get_color().g`` is the green channel, from 0 - 255, that is returned from RVR's color sensor.

        ``get_color().b`` is the blue channel, from 0 - 255, that is returned from RVR's color sensor."""
        if 'color_detection' in self.__sensor_data:
            color = self.__sensor_data['color_detection']
            return Color(round(color['r']), round(color['g']), round(color['b']))
        return None

    # BB-9E Sensors
    def get_dome_leds(self):
        """Provides the brightness of the Dome LEDs, from 0 to 15."""
        return self.__leds.get('dome', None)

    # R2-D2 & R2-Q5 Sensors
    def get_holo_projector_led(self):
        """Provides the brightness of the Holographic Projector LED, from 0 to 255."""
        return self.__leds.get('holo_projector', None)

    def get_logic_display_leds(self):
        """Provides the brightness of the white Logic Display LEDs, from 0 to 255."""
        return self.__leds.get('logic_display', None)

    # Communications
    def start_ir_broadcast(self, near: int, far: int):
        """Sets the IR emitters to broadcast on two specified channels, from 0 to 7, so other BOLTs can follow or evade.
        The broadcaster uses two channels because the first channel emits near IR pulses (< 1 meter), and the second
        channel emits far IR pulses (1 to 3 meters) so the following and evading BOLTs can detect these messages on
        their IR receivers with a sense of relative proximity to the broadcaster. You can't use a channel for more than
        one purpose at time, such as sending messages along with broadcasting, following, or evading. For example,
        use ``start_ir_broadcast(0, 1)`` to broadcast on channels 0 and 1, so that other BOLTs following or evading on
        0 and 1 will recognize this robot."""
        ToyUtil.start_robot_to_robot_infrared_broadcasting(self.__toy, bound_value(0, far, 7), bound_value(0, near, 7))

    def stop_ir_broadcast(self):
        """Stops the broadcasting behavior."""
        ToyUtil.stop_robot_to_robot_infrared_broadcasting(self.__toy)

    def start_ir_follow(self, near: int, far: int):
        """Sets the IR receivers to look for broadcasting BOLTs on the same channel pair, from 0 to 7. Upon receiving
        messages from a broadcasting BOLT, the follower will adjust its heading and speed to follow the broadcaster.
        When a follower loses sight of a broadcaster, the follower will spin in place to search for the broadcaster.
        You can't use a channel for more than one purpose at time, such as sending messages along with broadcasting,
        following, or evading. For example, use ``start_ir_follow(0, 1)`` to follow another BOLT that is broadcasting on
        channels 0 and 1."""
        ToyUtil.start_robot_to_robot_infrared_following(self.__toy, bound_value(0, far, 7), bound_value(0, near, 7))

    def stop_ir_follow(self):
        """Stops the following behavior."""
        ToyUtil.stop_robot_to_robot_infrared_following(self.__toy)

    def start_ir_evade(self, near: int, far: int):
        """Sets the IR receivers to look for broadcasting BOLTs on the same channel pair, from 0 to 7. Upon receiving
        messages from a broadcasting BOLT, the evader will adjust its heading to roll away from the broadcaster.
        When an evader loses sight of a broadcaster, the evader will spin in place to search for the broadcaster.
        The evader may stop if it is in the far range for a period of time so it does not roll too far away from the
        broadcaster. You can't use a channel for more than one purpose at time, such as sending messages along with
        broadcasting, following, or evading. For example, use ``start_ir_evade(0, 1)`` to evade another BOLT that is
        broadcasting on channels 0 and 1."""
        ToyUtil.start_robot_to_robot_infrared_evading(self.__toy, bound_value(0, far, 7), bound_value(0, near, 7))

    def stop_ir_evade(self):
        """Stops the evading behavior."""
        ToyUtil.stop_robot_to_robot_infrared_evading(self.__toy)

    def send_ir_message(self, channel: int, intensity: int):
        """Sends a message on a given IR channel, at a set intensity, from 1 to 64. Intensity is proportional to
        proximity, where a 1 is the closest, and 64 is the farthest. For example, use ``send_ir_message(4, 5)`` to send
        message 4 at intensity 5. You will need to use ``onIRMessage4(channel)`` event for on a corresponding robot to
        receive the message. Also see the ``getLastIRMessage()`` sensor to keep track of the last message your robot
        received. You can't use a channel for more than one purpose at time, such as sending messages along with
        broadcasting, following, or evading."""
        ToyUtil.send_robot_to_robot_infrared_message(
            self.__toy, bound_value(0, channel, 7), bound_value(1, intensity, 64))

    def listen_for_ir_message(self, channels: Union[int, Iterable[int]], duration: int = 0xFFFFFFFF):
        if isinstance(channels, int):
            channels = (channels,)
        if len(channels) > 0:
            ToyUtil.listen_for_robot_to_robot_infrared_message(
                self.__toy, map(lambda v: bound_value(0, v, 7), channels), bound_value(0, duration, 0xFFFFFFFF))

    def _robot_to_robot_infrared_message_received_notify(self, infrared_code: int):
        self.__last_message = infrared_code
        self.__call_event_listener(EventType.on_ir_message, infrared_code)

    def listen_for_color_sensor(self, colors: Iterable[Color]):
        if self.__toy.implements(IO.set_active_color_palette):
            palette = []
            for i, color in enumerate(colors):
                palette.extend((i, color.r, color.g, color.b))
            if palette:
                self.__toy.set_active_color_palette(palette)

    # Events: are predefined robot functions into which you can embed conditional logic. When an event occurs, the
    # conditional logic is called and then the program returns to the main loop where it left off. The event will
    # be called every time it occurs by default, unless you customize it.
    def __call_event_listener(self, event_type: EventType, *args, **kwargs):
        for f in self.__listeners[event_type]:
            threading.Thread(target=f, args=(self, *args), kwargs=kwargs).start()

    def register_event(self, event_type: EventType, listener: Callable[..., None]):
        """Registers the event type with listener. If listener is ``None`` then it removes all listeners of the
        specified event type.

        **Note**: listeners will be called in a newly spawned thread, meaning the caller have to deal with concurrency
        if needed. This library is thread-safe."""
        if event_type not in EventType:
            raise ValueError('Event type {event_type} does not exist')
        if listener:
            self.__listeners[event_type].add(listener)
        else:
            del self.__listeners[event_type]



from collections import OrderedDict
from enum import IntEnum
from functools import partialmethod

from spherov2.commands.animatronic import Animatronic
from spherov2.commands.drive import Drive
from spherov2.commands.io import IO
from spherov2.commands.power import Power
from spherov2.toy import ToySensor, Toy
from spherov2.toy.bb9e import BB9E
from spherov2.types import ToyType


class R2D2(BB9E):
    toy_type = ToyType('R2-D2', 'D2-', 'D2', .12)

    class LEDs(IntEnum):
        FRONT_RED = 0
        FRONT_GREEN = 1
        FRONT_BLUE = 2
        LOGIC_DISPLAYS = 3
        BACK_RED = 4
        BACK_GREEN = 5
        BACK_BLUE = 6
        HOLO_PROJECTOR = 7

    class Audio(IntEnum):
        TEST_1497HZ = 1
        TEST_200HZ = 32
        TEST_2517HZ = 63
        TEST_3581HZ = 94
        TEST_431HZ = 125
        TEST_6011HZ = 156
        TEST_853HZ = 187
        BB8_ALARM_1 = 218
        BB8_ALARM_10 = 235
        BB8_ALARM_11 = 254
        BB8_ALARM_12 = 258
        BB8_ALARM_2 = 264
        BB8_ALARM_3 = 268
        BB8_ALARM_4 = 272
        BB8_ALARM_6 = 279
        BB8_ALARM_7 = 288
        BB8_ALARM_8 = 296
        BB8_ALARM_9 = 301
        BB8_BOOT_UP = 309
        BB8_BOOR_UP_2 = 330
        BB8_CHATTY_1 = 352
        BB8_CHATTY_10 = 356
        BB8_CHATTY_11 = 360
        BB8_CHATTY_12 = 368
        BB8_CHATTY_13 = 375
        BB8_CHATTY_14 = 381
        BB8_CHATTY_15 = 390
        BB8_CHATTY_16 = 397
        BB8_CHATTY_17 = 404
        BB8_CHATTY_18 = 410
        BB8_CHATTY_19 = 417
        BB8_CHATTY_2 = 430
        BB8_CHATTY_20 = 464
        BB8_CHATTY_22 = 471
        BB8_CHATTY_23 = 479
        BB8_CHATTY_24 = 492
        BB8_CHATTY_25 = 518
        BB8_CHATTY_26 = 525
        BB8_CHATTY_27 = 537
        BB8_CHATTY_3 = 544
        BB8_CHATTY_4 = 557
        BB8_CHATTY_5 = 570
        BB8_CHATTY_6 = 577
        BB8_CHATTY_7 = 581
        BB8_CHATTY_8 = 587
        BB8_CHATTY_9 = 599
        BB8_DONT_KNOW = 606
        BB8_EXCITED_1 = 612
        BB8_EXCITED_2 = 622
        BB8_EXCITED_3 = 630
        BB8_EXCITED_4 = 644
        BB8_HEY_1 = 671
        BB8_HEY_10 = 682
        BB8_HEY_11 = 686
        BB8_HEY_12 = 690
        BB8_HEY_13 = 700
        BB8_HEY_2 = 724
        BB8_HEY_3 = 732
        BB8_HEY_4 = 734
        BB8_HEY_5 = 739
        BB8_HEY_6 = 743
        BB8_HEY_7 = 747
        BB8_HEY_8 = 753
        BB8_HEY_9 = 757
        BB8_LAUGH_1 = 761
        BB8_LAUGH_2 = 764
        BB8_NEGATIVE_1 = 787
        BB8_NEGATIVE_10 = 792
        BB8_NEGATIVE_11 = 802
        BB8_NEGATIVE_12 = 813
        BB8_NEGATIVE_13 = 825
        BB8_NEGATIVE_14 = 831
        BB8_NEGATIVE_15 = 836
        BB8_NEGATIVE_16 = 859
        BB8_NEGATIVE_17 = 867
        BB8_NEGATIVE_18 = 874
        BB8_NEGATIVE_19 = 888
        BB8_NEGATIVE_2 = 896
        BB8_NEGATIVE_20 = 902
        BB8_NEGATIVE_21 = 916
        BB8_NEGATIVE_22 = 927
        BB8_NEGATIVE_23 = 935
        BB8_NEGATIVE_24 = 946
        BB8_NEGATIVE_25 = 953
        BB8_NEGATIVE_26 = 963
        BB8_NEGATIVE_27 = 969
        BB8_NEGATIVE_28 = 983
        BB8_NEGATIVE_29 = 988
        BB8_NEGATIVE_3 = 998
        BB8_NEGATIVE_30 = 1003
        BB8_NEGATIVE_4 = 1010
        BB8_NEGATIVE_5 = 1020
        BB8_NEGATIVE_6 = 1032
        BB8_NEGATIVE_7 = 1040
        BB8_NEGATIVE_8 = 1052
        BB8_NEGATIVE_9 = 1064
        BB8_POSITIVE_1 = 1077
        BB8_POSITIVE_10 = 1082
        BB8_POSITIVE_11 = 1087
        BB8_POSITIVE_12 = 1092
        BB8_POSITIVE_13 = 1096
        BB8_POSITIVE_14 = 1102
        BB8_POSITIVE_15 = 1109
        BB8_POSITIVE_16 = 1113
        BB8_POSITIVE_2 = 1118
        BB8_POSITIVE_3 = 1123
        BB8_POSITIVE_4 = 1130
        BB8_POSITIVE_5 = 1135
        BB8_POSITIVE_6 = 1138
        BB8_POSITIVE_7 = 1147
        BB8_POSITIVE_8 = 1150
        BB8_POSITIVE_9 = 1157
        BB8_SAD_1 = 1163
        BB8_SAD_10 = 1170
        BB8_SAD_11 = 1178
        BB8_SAD_12 = 1186
        BB8_SAD_13 = 1192
        BB8_SAD_14 = 1199
        BB8_SAD_15 = 1205
        BB8_SAD_16 = 1213
        BB8_SAD_17 = 1220
        BB8_SAD_18 = 1230
        BB8_SAD_2 = 1236
        BB8_SAD_3 = 1240
        BB8_SAD_4 = 1250
        BB8_SAD_5 = 1268
        BB8_SAD_6 = 1275
        BB8_SAD_7 = 1281
        BB8_SAD_8 = 1295
        BB8_SAD_9 = 1303
        BB8_SHORTCUT = 1308
        BB8_WOW_1 = 1324
        BB9E_ALARM_1 = 1329
        BB9E_ALARM_2 = 1347
        BB9E_ALARM_3 = 1363
        BB9E_ALARM_4 = 1375
        BB9E_ALARM_5 = 1384
        BB9E_CHATTY_1 = 1394
        BB9E_CHATTY_2 = 1408
        BB9E_EXCITED_1 = 1432
        BB9E_EXCITED_2 = 1442
        BB9E_EXCITED_3 = 1463
        BB9E_HEY_1 = 1476
        BB9E_HEY_2 = 1486
        BB9E_NEGATIVE_1 = 1494
        BB9E_NEGATIVE_2 = 1505
        BB9E_NEGATIVE_3 = 1516
        BB9E_NEGATIVE_4 = 1523
        BB9E_POSITIVE_1 = 1531
        BB9E_POSITIVE_2 = 1549
        BB9E_POSITIVE_3 = 1558
        BB9E_POSITIVE_4 = 1571
        BB9E_POSITIVE_5 = 1583
        BB9E_SAD_1 = 1592
        BB9E_SAD_2 = 1600
        R2_FALL = 1609
        R2_HIT_1 = 1623
        R2_HIT_10 = 1628
        R2_HIT_11 = 1635
        R2_HIT_2 = 1642
        R2_HIT_3 = 1647
        R2_HIT_4 = 1653
        R2_HIT_5 = 1659
        R2_HIT_6 = 1664
        R2_HIT_7 = 1669
        R2_HIT_8 = 1676
        R2_HIT_9 = 1684
        R2_STEP_1 = 1690
        R2_STEP_2 = 1693
        R2_STEP_3 = 1696
        R2_STEP_4 = 1698
        R2_STEP_5 = 1700
        R2_STEP_6 = 1702
        R2_ACCESS_PANELS = 1704
        R2_ALARM_1 = 1737
        R2_ALARM_10 = 1747
        R2_ALARM_12 = 1756
        R2_ALARM_13 = 1763
        R2_ALARM_14 = 1771
        R2_ALARM_15 = 1784
        R2_ALARM_16 = 1791
        R2_ALARM_2 = 1809
        R2_ALARM_3 = 1821
        R2_ALARM_4 = 1831
        R2_ALARM_5 = 1835
        R2_ALARM_6 = 1843
        R2_ALARM_7 = 1858
        R2_ALARM_8 = 1867
        R2_ALARM_9 = 1893
        R2_ANNOYED = 1910
        R2_BURNOUT = 1915
        R2_CHATTY_1 = 1950
        R2_CHATTY_10 = 1959
        R2_CHATTY_11 = 1966
        R2_CHATTY_12 = 1977
        R2_CHATTY_13 = 1987
        R2_CHATTY_14 = 2002
        R2_CHATTY_15 = 2007
        R2_CHATTY_16 = 2010
        R2_CHATTY_17 = 2019
        R2_CHATTY_18 = 2028
        R2_CHATTY_19 = 2039
        R2_CHATTY_2 = 2061
        R2_CHATTY_20 = 2072
        R2_CHATTY_21 = 2080
        R2_CHATTY_22 = 2085
        R2_CHATTY_23 = 2095
        R2_CHATTY_24 = 2105
        R2_CHATTY_25 = 2121
        R2_CHATTY_26 = 2132
        R2_CHATTY_27 = 2143
        R2_CHATTY_28 = 2157
        R2_CHATTY_29 = 2170
        R2_CHATTY_3 = 2174
        R2_CHATTY_30 = 2184
        R2_CHATTY_31 = 2188
        R2_CHATTY_32 = 2198
        R2_CHATTY_33 = 2202
        R2_CHATTY_34 = 2211
        R2_CHATTY_35 = 2221
        R2_CHATTY_36 = 2232
        R2_CHATTY_37 = 2241
        R2_CHATTY_38 = 2253
        R2_CHATTY_39 = 2264
        R2_CHATTY_4 = 2276
        R2_CHATTY_40 = 2285
        R2_CHATTY_41 = 2292
        R2_CHATTY_42 = 2307
        R2_CHATTY_43 = 2322
        R2_CHATTY_44 = 2332
        R2_CHATTY_45 = 2344
        R2_CHATTY_46 = 2357
        R2_CHATTY_47 = 2368
        R2_CHATTY_48 = 2377
        R2_CHATTY_49 = 2387
        R2_CHATTY_5 = 2399
        R2_CHATTY_50 = 2413
        R2_CHATTY_51 = 2424
        R2_CHATTY_52 = 2439
        R2_CHATTY_53 = 2452
        R2_CHATTY_54 = 2457
        R2_CHATTY_55 = 2463
        R2_CHATTY_56 = 2474
        R2_CHATTY_57 = 2492
        R2_CHATTY_58 = 2509
        R2_CHATTY_59 = 2519
        R2_CHATTY_6 = 2524
        R2_CHATTY_60 = 2535
        R2_CHATTY_61 = 2543
        R2_CHATTY_62 = 2554
        R2_CHATTY_7 = 2562
        R2_CHATTY_8 = 2572
        R2_CHATTY_9 = 2579
        R2_ENGAGE_HYPER_DRIVE = 2586
        R2_EXCITED_1 = 2600
        R2_EXCITED_10 = 2615
        R2_EXCITED_11 = 2633
        R2_EXCITED_12 = 2644
        R2_EXCITED_13 = 2654
        R2_EXCITED_14 = 2662
        R2_EXCITED_15 = 2680
        R2_EXCITED_16 = 2691
        R2_EXCITED_2 = 2708
        R2_EXCITED_3 = 2726
        R2_EXCITED_4 = 2730
        R2_EXCITED_5 = 2736
        R2_EXCITED_6 = 2753
        R2_EXCITED_7 = 2767
        R2_EXCITED_8 = 2777
        R2_EXCITED_9 = 2787
        R2_HEAD_SPIN = 2797
        R2_HEY_1 = 2813
        R2_HEY_10 = 2824
        R2_HEY_11 = 2828
        R2_HEY_12 = 2833
        R2_HEY_2 = 2841
        R2_HEY_3 = 2856
        R2_HEY_4 = 2861
        R2_HEY_5 = 2882
        R2_HEY_6 = 2893
        R2_HEY_7 = 2898
        R2_HEY_8 = 2904
        R2_HEY_9 = 2912
        R2_LAUGH_1 = 2919
        R2_LAUGH_2 = 2935
        R2_LAUGH_3 = 2950
        R2_LAUGH_4 = 2955
        R2_MOTOR = 2970
        R2_NEGATIVE_1 = 3101
        R2_NEGATIVE_10 = 3111
        R2_NEGATIVE_11 = 3115
        R2_NEGATIVE_12 = 3121
        R2_NEGATIVE_13 = 3132
        R2_NEGATIVE_14 = 3136
        R2_NEGATIVE_15 = 3148
        R2_NEGATIVE_16 = 3152
        R2_NEGATIVE_17 = 3157
        R2_NEGATIVE_18 = 3164
        R2_NEGATIVE_19 = 3167
        R2_NEGATIVE_2 = 3172
        R2_NEGATIVE_20 = 3178
        R2_NEGATIVE_21 = 3191
        R2_NEGATIVE_22 = 3200
        R2_NEGATIVE_23 = 3213
        R2_NEGATIVE_24 = 3219
        R2_NEGATIVE_25 = 3226
        R2_NEGATIVE_26 = 3230
        R2_NEGATIVE_27 = 3233
        R2_NEGATIVE_28 = 3241
        R2_NEGATIVE_3 = 3251
        R2_NEGATIVE_4 = 3258
        R2_NEGATIVE_5 = 3263
        R2_NEGATIVE_6 = 3268
        R2_NEGATIVE_7 = 3274
        R2_NEGATIVE_8 = 3282
        R2_NEGATIVE_9 = 3291
        R2_POSITIVE_1 = 3302
        R2_POSITIVE_10 = 3309
        R2_POSITIVE_11 = 3318
        R2_POSITIVE_12 = 3326
        R2_POSITIVE_13 = 3340
        R2_POSITIVE_14 = 3353
        R2_POSITIVE_15 = 3358
        R2_POSITIVE_16 = 3364
        R2_POSITIVE_17 = 3369
        R2_POSITIVE_18 = 3375
        R2_POSITIVE_19 = 3388
        R2_POSITIVE_2 = 3394
        R2_POSITIVE_20 = 3403
        R2_POSITIVE_21 = 3410
        R2_POSITIVE_22 = 3422
        R2_POSITIVE_23 = 3434
        R2_POSITIVE_3 = 3439
        R2_POSITIVE_4 = 3446
        R2_POSITIVE_5 = 3449
        R2_POSITIVE_6 = 3454
        R2_POSITIVE_7 = 3460
        R2_POSITIVE_8 = 3471
        R2_POSITIVE_9 = 3478
        R2_SAD_1 = 3484
        R2_SAD_10 = 3495
        R2_SAD_11 = 3518
        R2_SAD_12 = 3526
        R2_SAD_13 = 3536
        R2_SAD_14 = 3543
        R2_SAD_15 = 3553
        R2_SAD_16 = 3561
        R2_SAD_17 = 3570
        R2_SAD_18 = 3593
        R2_SAD_19 = 3600
        R2_SAD_2 = 3608
        R2_SAD_20 = 3612
        R2_SAD_21 = 3619
        R2_SAD_22 = 3632
        R2_SAD_23 = 3639
        R2_SAD_24 = 3649
        R2_SAD_25 = 3661
        R2_SAD_3 = 3686
        R2_SAD_4 = 3693
        R2_SAD_5 = 3703
        R2_SAD_6 = 3739
        R2_SAD_7 = 3755
        R2_SAD_8 = 3782
        R2_SAD_9 = 3790
        R2_SCREAM = 3797
        R2_SCREAM_2 = 3810
        R2_SHORT_OUT = 3825
        R2Q5_ALARM_1 = 3864
        R2Q5_ALARM_2 = 3869
        R2Q5_CHATTY_1 = 3875
        R2Q5_CHATTY_2 = 3891
        R2Q5_HEY_1 = 3896
        R2Q5_HEY_2 = 3901
        R2Q5_NEGATIVE_1 = 3907
        R2Q5_POSITIVE_1 = 3914
        R2Q5_POSITIVE_2 = 3920
        R2Q5_SAD_1 = 3925
        R2Q5_SHUTDOWN = 3929
        BB9E_EXTRA_1 = 3941
        BB9E_EXTRA_2 = 4021
        BB9E_EXTRA_3 = 4246
        BB9E_EXTRA_4 = 4568
        BB9E_EXTRA_5 = 4929
        BB9E_EXTRA_6 = 5156
        BB9E_EXTRA_7 = 5315
        BB9E_HEAD_TURN_1 = 5441
        BB9E_HEAD_TURN_2 = 5483
        BB9E_HEAD_TURN_3 = 5513

    class Animations(IntEnum):
        CHARGER_1 = 0
        CHARGER_2 = 1
        CHARGER_3 = 2
        CHARGER_4 = 3
        CHARGER_5 = 4
        CHARGER_6 = 5
        CHARGER_7 = 6
        EMOTE_ALARM = 7
        EMOTE_ANGRY = 8
        EMOTE_ATTENTION = 9
        EMOTE_FRUSTRATED = 10
        EMOTE_DRIVE = 11
        EMOTE_EXCITED = 12
        EMOTE_SEARCH = 13
        EMOTE_SHORT_CIRCUIT = 14
        EMOTE_LAUGH = 15
        EMOTE_NO = 16
        EMOTE_RETREAT = 17
        EMOTE_FIERY = 18
        EMOTE_UNDERSTOOD = 19
        EMOTE_YES = 21
        EMOTE_SCAN = 22
        EMOTE_SURPRISED = 24
        IDLE_1 = 25
        IDLE_2 = 26
        IDLE_3 = 27
        WWM_ANGRY = 31
        WWM_ANXIOUS = 32
        WWM_BOW = 33
        WWM_CONCERN = 34
        WWM_CURIOUS = 35
        WWM_DOUBLE_TAKE = 36
        WWM_EXCITED = 37
        WWM_FIERY = 38
        WMM_FRUSTRATED = 39
        WWM_HAPPY = 40
        WWM_JITTERY = 41
        WWM_LAUGH = 42
        WWM_LONG_SHAKE = 43
        WWM_NO = 44
        WWM_OMINOUS = 45
        WWM_RELIEVED = 46
        WWM_SAD = 47
        WWM_SCARED = 48
        WWM_SHAKE = 49
        WWM_SURPRISED = 50
        WWM_TAUNTING = 51
        WWM_WHISPER = 52
        WWM_YELLING = 53
        WWM_YOOHOO = 54
        MOTOR = 55

    extended_sensors = OrderedDict(
        r2_head_angle=OrderedDict(r2_head_angle=ToySensor(0x4000000, -162., 182.)),
        gyroscope=OrderedDict(
            x=ToySensor(0x2000000, -20000., 20000.),
            y=ToySensor(0x1000000, -20000., 20000.),
            z=ToySensor(0x800000, -20000., 20000.)
        )
    )

    get_battery_voltage_state = Power.get_battery_voltage_state

    generic_raw_motor = Drive.generic_raw_motor

    perform_leg_action = Animatronic.perform_leg_action
    set_head_position = Animatronic.set_head_position
    get_head_position = Animatronic.get_head_position
    set_leg_position = Animatronic.set_leg_position
    get_leg_position = Animatronic.get_leg_position
    get_leg_action = Animatronic.get_leg_action
    enable_leg_action_notify = Animatronic.enable_leg_action_notify
    enable_head_reset_to_zero_notify = Animatronic.enable_head_reset_to_zero_notify
    add_head_reset_to_zero_notify_listener = partialmethod(Toy._add_listener, Animatronic.head_reset_to_zero_notify)
    remove_head_reset_to_zero_notify_listener = partialmethod(Toy._remove_listener,
                                                              Animatronic.head_reset_to_zero_notify)

    set_audio_volume = IO.set_audio_volume
    get_audio_volume = IO.get_audio_volume
    stop_all_audio = IO.stop_all_audio



from enum import IntEnum
from typing import NamedTuple


class CollisionDetected(NamedTuple):
    acceleration_x: float
    acceleration_y: float
    acceleration_z: float
    x_axis: bool
    y_axis: bool
    power_x: int
    power_y: int
    power_z: int
    speed: int
    time: float


class SensorStreamingMask(NamedTuple):
    interval: int
    packet_count: int
    data_mask: int


class BotToBotInfraredReadings(NamedTuple):
    back_left: bool
    back_right: bool
    front_left: bool
    front_right: bool


class RgbcSensorValues(NamedTuple):
    red: int
    green: int
    blue: int
    clear: int


class ColorDetection(NamedTuple):
    red: int
    green: int
    blue: int
    confidence: int
    color_classification_id: int


class StreamingServiceData(NamedTuple):
    token: int
    sensor_data: bytearray


class MotorCurrent(NamedTuple):
    left_motor_current: float
    right_motor_current: float
    up_time: int


class MotorTemperature(NamedTuple):
    case_temperature: float
    winding_coil_temperature: float


class ThermalProtectionStatus(IntEnum):
    OK = 0
    WARN = 1
    CRITICAL = 2


class MotorThermalProtectionStatus(NamedTuple):
    left_motor_temperature: float
    left_motor_status: ThermalProtectionStatus
    right_motor_temperature: float
    right_motor_status: ThermalProtectionStatus



from spherov2.helper import nibble_to_byte


class Commands:
    @classmethod
    def _encode(cls, toy, cid, proc, data=None):
        if toy._require_target and not proc:
            raise ValueError(f'Robot type {cls} requires target processor')
        if proc:
            proc = nibble_to_byte(1, proc)
        return toy._packet_manager.new_packet(cls._did, cid, proc, data)



import struct
from enum import IntEnum

from spherov2.commands import Commands
from spherov2.helper import to_bytes, to_int
from spherov2.listeners.core import Versions, BluetoothInfo, PowerState, PowerStates


class IntervalOptions(IntEnum):
    NONE = 0
    DEEP_SLEEP = 0xffff


class ChargerStates(IntEnum):
    OUT = 0
    IN = 1
    UNKNOWN = 2


class PowerStates(IntEnum):
    UNKNOWN = 0
    CHARGING = 1
    OK = 2
    LOW = 3
    CRITICAL = 4


class Core(Commands):
    _did = 0

    @staticmethod
    def ping(toy, proc=None):
        return Core._encode(toy, 1, proc)

    @staticmethod
    def get_versions(toy, proc=None):
        unpacked = struct.unpack('>8B', toy._execute(Core._encode(toy, 2, proc)).data)
        return Versions(
            record_version=unpacked[0], model_number=unpacked[1], hardware_version_code=unpacked[2],
            main_app_version_major=unpacked[3], main_app_version_minor=unpacked[4],
            bootloader_version='%d.%d' % (unpacked[5] >> 4, unpacked[5] & 0xf),
            orb_basic_version='%d.%d' % (unpacked[6] >> 4, unpacked[6] & 0xf),
            overlay_version='%d.%d' % (unpacked[7] >> 4, unpacked[7] & 0xf),
        )

    @staticmethod
    def set_bluetooth_name(toy, name: bytes, proc=None):
        toy._execute(Core._encode(toy, 16, proc, [*name, 0]))

    @staticmethod
    def get_bluetooth_info(toy, proc=None):
        data = toy._execute(Core._encode(toy, 17, proc)).data.rstrip(b'\0')
        name, *_, address = data.split(b'\0')
        return BluetoothInfo(bytes(name), bytes(address))

    @staticmethod
    def get_power_state(toy, proc=None):
        unpacked = struct.unpack('>2B3H', toy._execute(Core._encode(toy, 32, proc)).data)
        return PowerState(record_version=unpacked[0], state=PowerStates(unpacked[1]), voltage=unpacked[2] / 100,
                          number_of_charges=unpacked[3], time_since_last_charge=unpacked[4])

    @staticmethod
    def enable_battery_state_changed_notify(toy, enable: bool, proc=None):
        toy._execute(Core._encode(toy, 33, proc, [int(enable)]))

    @staticmethod
    def sleep(toy, interval_option: IntervalOptions, unk: int, unk2: int, proc=None):
        return toy._execute(Core._encode(toy, 34, proc, [*to_bytes(interval_option, 2), unk, *to_bytes(unk2, 2)]))

    @staticmethod
    def set_inactivity_timeout(toy, timeout: int, proc=None):
        toy._execute(Core._encode(toy, 37, proc, to_bytes(timeout, 2)))

    @staticmethod
    def get_charger_state(toy, proc=None):
        return ChargerStates(toy._execute(Core._encode(toy, 38, proc)).data[0])

    @staticmethod
    def get_factory_config_block_crc(toy, proc=None):
        return to_int(toy._execute(Core._encode(toy, 39, proc)).data)

    @staticmethod
    def jump_to_bootloader(toy, proc=None):
        toy._execute(Core._encode(toy, 48, proc))



import threading
from collections import OrderedDict, defaultdict
from enum import IntEnum, Enum, auto, IntFlag
from typing import Dict, List, Callable, NamedTuple, Tuple

from spherov2.commands.drive import DriveFlags
from spherov2.commands.drive import RawMotorModes as DriveRawMotorModes
from spherov2.commands.io import IO
from spherov2.controls import RawMotorModes, PacketDecodingException, CommandExecuteError
from spherov2.helper import to_bytes, to_int, packet_chk
from spherov2.listeners.sensor import StreamingServiceData


class Packet(NamedTuple):
    """Packet protocol v2, from https://sdk.sphero.com/docs/api_spec/general_api
    [SOP, FLAGS, TID (optional), SID (optional), DID, CID, SEQ, ERR (at response), DATA..., CHK, EOP]"""

    flags: 'Packet.Flags'
    did: int
    cid: int
    seq: int
    tid: int
    sid: int
    data: bytearray
    err: 'Packet.Error' = None

    class Flags(IntFlag):
        is_response = 0b1
        requests_response = 0b10
        requests_only_error_response = 0b100
        is_activity = 0b1000
        has_target_id = 0b10000
        has_source_id = 0b100000
        unused = 0b1000000
        extended_flags = 0b10000000

    class Encoding(IntEnum):
        escape = 0xAB
        start = 0x8D
        end = 0xD8
        escaped_escape = 0x23
        escaped_start = 0x05
        escaped_end = 0x50

    class Error(IntEnum):
        success = 0x00
        bad_device_id = 0x01
        bad_command_id = 0x02
        not_yet_implemented = 0x03
        command_is_restricted = 0x04
        bad_data_length = 0x05
        command_failed = 0x06
        bad_parameter_value = 0x07
        busy = 0x08
        bad_target_id = 0x09
        target_unavailable = 0x0a

    @staticmethod
    def parse_response(data) -> 'Packet':
        sop, *data, eop = data
        if sop != Packet.Encoding.start:
            raise PacketDecodingException('Unexpected start of packet')
        if eop != Packet.Encoding.end:
            raise PacketDecodingException('Unexpected end of packet')
        *data, chk = Packet.__unescape_data(data)
        if packet_chk(data) != chk:
            raise PacketDecodingException('Bad response checksum')

        flags = data.pop(0)

        tid = None
        if flags & Packet.Flags.has_target_id:
            tid = data.pop(0)

        sid = None
        if flags & Packet.Flags.has_source_id:
            sid = data.pop(0)

        did, cid, seq, *data = data

        err = Packet.Error.success
        if flags & Packet.Flags.is_response:
            err = Packet.Error(data.pop(0))

        return Packet(flags, did, cid, seq, tid, sid, bytearray(data), err)

    @staticmethod
    def __unescape_data(response_data) -> List[int]:
        raw_data = []

        iter_response_data = iter(response_data)
        for b in iter_response_data:
            if b == Packet.Encoding.escape:
                b = next(iter_response_data, None)
                if b == Packet.Encoding.escaped_escape:
                    b = Packet.Encoding.escape
                elif b == Packet.Encoding.escaped_start:
                    b = Packet.Encoding.start
                elif b == Packet.Encoding.escaped_end:
                    b = Packet.Encoding.end
                else:
                    raise PacketDecodingException('Unexpected escaping byte')

            raw_data.append(b)

        return raw_data

    @property
    def id(self) -> Tuple:
        return self.did, self.cid, self.seq

    def build(self) -> bytearray:
        packet = bytearray([self.flags])

        if self.flags & Packet.Flags.has_target_id:
            packet.append(self.tid)

        if self.flags & Packet.Flags.has_source_id:
            packet.append(self.sid)

        packet.extend(self.id)

        if self.flags & Packet.Flags.is_response:
            packet.append(self.err)

        packet.extend(self.data)
        packet.append(packet_chk(packet))

        escaped_packet = bytearray([Packet.Encoding.start])
        for c in packet:
            if c == Packet.Encoding.escape:
                escaped_packet.extend((Packet.Encoding.escape, Packet.Encoding.escaped_escape))
            elif c == Packet.Encoding.start:
                escaped_packet.extend((Packet.Encoding.escape, Packet.Encoding.escaped_start))
            elif c == Packet.Encoding.end:
                escaped_packet.extend((Packet.Encoding.escape, Packet.Encoding.escaped_end))
            else:
                escaped_packet.append(c)
        escaped_packet.append(Packet.Encoding.end)

        return escaped_packet

    def check_error(self):
        if self.err != Packet.Error.success:
            raise CommandExecuteError(self.err)

    class Manager:
        def __init__(self):
            self.__seq = 0

        def new_packet(self, did, cid, tid=None, data=None):
            flags = Packet.Flags.requests_response | Packet.Flags.is_activity
            sid = None
            if tid is not None:
                flags |= Packet.Flags.has_source_id | Packet.Flags.has_target_id
                sid = 0x1
            packet = Packet(flags, did, cid, self.__seq, tid, sid, bytearray(data or []))
            self.__seq = (self.__seq + 1) % 0xff
            return packet

    class Collector:
        def __init__(self, callback):
            self.__callback = callback
            self.__data = []

        def add(self, data):
            for b in data:
                self.__data.append(b)
                if b == Packet.Encoding.end:
                    pkt = self.__data
                    self.__data = []
                    if len(pkt) < 6:
                        raise PacketDecodingException(f'Very small packet {[hex(x) for x in pkt]}')
                    self.__callback(Packet.parse_response(pkt))


class AnimationControl:
    def __init__(self, toy):
        self.__toy = toy


class DriveControl:
    def __init__(self, toy):
        self.__toy = toy
        self.__is_boosting = False

    def roll_start(self, heading, speed):
        flag = DriveFlags.FORWARD
        if speed < 0:
            flag = DriveFlags.BACKWARD
            heading = (heading + 180) % 360
        if self.__is_boosting:
            flag |= DriveFlags.TURBO
        speed = min(255, abs(speed))
        self.__toy.drive_with_heading(speed, heading, flag)

    def roll_stop(self, heading):
        self.roll_start(heading, 0)

    set_heading = roll_stop

    def set_stabilization(self, stabilize):
        self.__toy.set_stabilization(stabilize)

    def set_raw_motors(self, left_mode, left_speed, right_mode, right_speed):
        if left_mode == RawMotorModes.FORWARD:
            left_drive_mode = DriveRawMotorModes.FORWARD
        elif left_mode == RawMotorModes.REVERSE:
            left_drive_mode = DriveRawMotorModes.REVERSE
        else:
            left_drive_mode = DriveRawMotorModes.OFF

        if right_mode == RawMotorModes.FORWARD:
            right_drive_mode = DriveRawMotorModes.FORWARD
        elif right_mode == RawMotorModes.REVERSE:
            right_drive_mode = DriveRawMotorModes.REVERSE
        else:
            right_drive_mode = DriveRawMotorModes.OFF

        self.__toy.set_raw_motors(left_drive_mode, left_speed, right_drive_mode, right_speed)

    def reset_heading(self):
        self.__toy.reset_yaw()


class FirmwareUpdateControl:
    def __init__(self, toy):
        self.__toy = toy


class LedControl:
    def __init__(self, toy):
        self.__toy = toy

    def set_leds(self, mapping: Dict[IntEnum, int]):
        mask = 0
        led_values = []
        for e in self.__toy.LEDs:
            if e in mapping:
                mask |= 1 << e
                led_values.append(mapping[e])
        if mask:
            if self.__toy.implements(IO.set_all_leds_with_32_bit_mask):
                self.__toy.set_all_leds_with_32_bit_mask(mask, led_values)
            elif self.__toy.implements(IO.set_all_leds_with_16_bit_mask):
                self.__toy.set_all_leds_with_16_bit_mask(mask, led_values)
            elif hasattr(self.__toy, 'set_all_leds_with_8_bit_mask'):
                self.__toy.set_all_leds_with_8_bit_mask(mask, led_values)


class SensorControl:
    def __init__(self, toy):
        toy.add_sensor_streaming_data_notify_listener(self.__process_sensor_stream_data)

        self.__toy = toy
        self.__count = 0
        self.__interval = 250
        self.__enabled = {}
        self.__enabled_extended = {}
        self.__listeners = set()

    def __process_sensor_stream_data(self, sensor_data: List[float]):
        data = {}

        def __new_data():
            n = {}
            for name, component in components.items():
                d = sensor_data.pop(0)
                if component.modifier:
                    d = component.modifier(d)
                n[name] = d
            data[sensor] = n

        for sensor, components in self.__toy.sensors.items():
            if sensor in self.__enabled:
                __new_data()
        for sensor, components in self.__toy.extended_sensors.items():
            if sensor in self.__enabled_extended:
                __new_data()

        for f in self.__listeners:
            threading.Thread(target=f, args=(data,)).start()

    def add_sensor_data_listener(self, listener: Callable[[Dict[str, Dict[str, float]]], None]):
        self.__listeners.add(listener)

    def remove_sensor_data_listener(self, listener: Callable[[Dict[str, Dict[str, float]]], None]):
        self.__listeners.remove(listener)

    def set_count(self, count: int):
        if count >= 0 and count != self.__count:
            self.__count = count
            self.__update()

    def set_interval(self, interval: int):
        if interval >= 0 and interval != self.__interval:
            self.__interval = interval
            self.__update()

    def __update(self):
        sensors_mask = extended_sensors_mask = 0
        for sensor in self.__enabled.values():
            for component in sensor.values():
                sensors_mask |= component.bit
        for sensor in self.__enabled_extended.values():
            for component in sensor.values():
                extended_sensors_mask |= component.bit
        self.__toy.set_sensor_streaming_mask(0, self.__count, sensors_mask)
        self.__toy.set_extended_sensor_streaming_mask(extended_sensors_mask)
        self.__toy.set_sensor_streaming_mask(self.__interval, self.__count, sensors_mask)

    def enable(self, *sensors):
        for sensor in sensors:
            if sensor in self.__toy.sensors:
                self.__enabled[sensor] = self.__toy.sensors[sensor]
            elif sensor in self.__toy.extended_sensors:
                self.__enabled_extended[sensor] = self.__toy.extended_sensors[sensor]
        self.__update()

    def disable(self, *sensors):
        for sensor in sensors:
            self.__enabled.pop(sensor, None)
            self.__enabled_extended.pop(sensor, None)
        self.__update()

    def disable_all(self):
        self.__enabled.clear()
        self.__enabled_extended.clear()
        self.__update()


class StatsControl:
    def __init__(self, toy):
        self.__toy = toy


class Processors(IntEnum):
    UNKNOWN = 0
    PRIMARY = 1
    SECONDARY = 2


class StreamingServiceAttribute(NamedTuple):
    min_value: int
    max_value: int
    modifier: Callable[[float], float] = None


class StreamingDataSizes(IntEnum):
    EightBit = 0
    SixteenBit = 1
    ThirtyTwoBit = 2


class StreamingService(NamedTuple):
    attributes: OrderedDict
    slot: int
    processor: Processors = Processors.SECONDARY
    data_size: StreamingDataSizes = StreamingDataSizes.ThirtyTwoBit


class StreamingServiceState(Enum):
    Unknown = auto()
    Stop = auto()
    Start = auto()
    Restart = auto()


class StreamingControl:
    __streaming_services = {
        'quaternion': StreamingService(OrderedDict(
            w=StreamingServiceAttribute(-1, 1),
            x=StreamingServiceAttribute(-1, 1),
            y=StreamingServiceAttribute(-1, 1),
            z=StreamingServiceAttribute(-1, 1)
        ), 1),
        'imu': StreamingService(OrderedDict(
            pitch=StreamingServiceAttribute(-180, 180),
            roll=StreamingServiceAttribute(-90, 90),
            yaw=StreamingServiceAttribute(-180, 180)
        ), 1),
        'accelerometer': StreamingService(OrderedDict(
            x=StreamingServiceAttribute(-16, 16),
            y=StreamingServiceAttribute(-16, 16),
            z=StreamingServiceAttribute(-16, 16)
        ), 1),
        'color_detection': StreamingService(OrderedDict(
            r=StreamingServiceAttribute(0, 255),
            g=StreamingServiceAttribute(0, 255),
            b=StreamingServiceAttribute(0, 255),
            index=StreamingServiceAttribute(0, 255),
            confidence=StreamingServiceAttribute(0, 1)
        ), 1, Processors.PRIMARY, StreamingDataSizes.EightBit),
        'gyroscope': StreamingService(OrderedDict(
            x=StreamingServiceAttribute(-2000, 2000),
            y=StreamingServiceAttribute(-2000, 2000),
            z=StreamingServiceAttribute(-2000, 2000)
        ), 1),
        'core_time_lower': StreamingService(OrderedDict(time_lower=StreamingServiceAttribute(0, 1 << 63)), 3),
        'locator': StreamingService(OrderedDict(
            x=StreamingServiceAttribute(-16000, 16000, lambda x: x * 100.),
            y=StreamingServiceAttribute(-16000, 16000, lambda x: x * 100.),
        ), 2),
        'velocity': StreamingService(OrderedDict(
            x=StreamingServiceAttribute(-5, 5, lambda x: x * 100.),
            y=StreamingServiceAttribute(-5, 5, lambda x: x * 100.),
        ), 2),
        'speed': StreamingService(OrderedDict(speed=StreamingServiceAttribute(0, 5, lambda x: x * 100.)), 2),
        'core_time_upper': StreamingService(OrderedDict(time_upper=StreamingServiceAttribute(0, 1 << 63)), 3),
        'ambient_light': StreamingService(
            OrderedDict(light=StreamingServiceAttribute(0, 120000)), 2, Processors.PRIMARY
        ),
    }

    def __init__(self, toy):
        toy.add_streaming_service_data_notify_listener(self.__streaming_service_data)
        self.__toy = toy
        self.__slots = {
            Processors.PRIMARY: defaultdict(list),
            Processors.SECONDARY: defaultdict(list)
        }
        self.__enabled = set()
        self.__listeners = set()
        self.__interval = 500

    def add_sensor_data_listener(self, listener: Callable[[Dict[str, Dict[str, float]]], None]):
        self.__listeners.add(listener)

    def remove_sensor_data_listener(self, listener: Callable[[Dict[str, Dict[str, float]]], None]):
        self.__listeners.remove(listener)

    def enable(self, *sensors):
        changed = False
        for sensor in sensors:
            if sensor not in self.__enabled and sensor in self.__streaming_services:
                self.__enabled.add(sensor)
                changed = True
        if changed:
            self.__configure(StreamingServiceState.Start)

    def disable(self, *sensors):
        changed = False
        for sensor in sensors:
            if sensor in self.__enabled:
                self.__enabled.remove(sensor)
                changed = True
        if changed:
            self.__configure(StreamingServiceState.Start if self.__enabled else StreamingServiceState.Stop)

    def disable_all(self):
        if not self.__enabled:
            return
        self.__enabled.clear()
        self.__configure(StreamingServiceState.Stop)

    def set_count(self, count: int):
        pass

    def set_interval(self, interval: int):
        if interval < 0:
            raise ValueError('Interval attempted to be set with negative value')
        self.__interval = interval
        if self.__enabled:
            self.__configure(StreamingServiceState.Restart)

    def __configure(self, state: StreamingServiceState):
        for target in [Processors.PRIMARY, Processors.SECONDARY]:
            self.__toy.stop_streaming_service(target)
            if state == StreamingServiceState.Stop:
                self.__toy.clear_streaming_service(target)
            elif state == StreamingServiceState.Start:
                self.__toy.clear_streaming_service(target)
                slots = self.__slots[target]
                slots.clear()
                for index, (s, sensor) in enumerate(self.__streaming_services.items()):
                    if s in self.__enabled and sensor.processor == target:
                        slots[sensor.slot].append((index, s, sensor))
                if slots:
                    for slot, services in slots.items():
                        data = []
                        for index, _, sensor in services:
                            data.extend(to_bytes(index, 2))
                            data.append(sensor.data_size)
                        self.__toy.configure_streaming_service(slot, data, target)
                    self.__toy.start_streaming_service(self.__interval, target)
            elif state == StreamingServiceState.Restart:
                self.__toy.start_streaming_service(self.__interval, target)

    def __streaming_service_data(self, source_id, data: StreamingServiceData):
        node = data.token & 0xf
        processor = source_id & 0xf
        sensor_data = data.sensor_data
        services = self.__slots[processor][node]
        data = {}
        for _, sensor_name, sensor in services:
            n = {}
            for name, component in sensor.attributes.items():
                data_size = 1 << sensor.data_size
                value, sensor_data = to_int(sensor_data[:data_size]), sensor_data[data_size:]
                value = value / ((1 << data_size * 8) - 1) * (
                        component.max_value - component.min_value) + component.min_value
                if component.modifier is not None:
                    value = component.modifier(value)
                n[name] = value
            if sensor_name == 'color_detection' and node != 0:
                continue
            data[sensor_name] = n
        for f in self.__listeners:
            threading.Thread(target=f, args=(data,)).start()



from spherov2.commands import Commands
from spherov2.listeners.api_and_shell import ApiProtocolVersion


class ApiAndShell(Commands):
    _did = 16

    @staticmethod
    def ping(toy, data, proc=None) -> bytearray:
        return toy._execute(ApiAndShell._encode(toy, 0, proc, data)).data

    @staticmethod
    def get_api_protocol_version(toy, proc=None) -> ApiProtocolVersion:
        data = toy._execute(ApiAndShell._encode(toy, 1, proc)).data
        return ApiProtocolVersion(major_version=data[0], minor_version=data[1])

    @staticmethod
    def send_command_to_shell(toy, command: bytes, proc=None):
        toy._execute(ApiAndShell._encode(toy, 2, proc, [*command, 0]))

    send_string_to_console = (16, 3, lambda listener, p: listener(bytes(p.data).rstrip(b'\0')))

    @staticmethod
    def get_supported_dids(toy, proc=None):
        return list(toy._execute(ApiAndShell._encode(toy, 5, proc)).data)

    @staticmethod
    def get_supported_cids(toy, s, proc=None):
        return list(toy._execute(ApiAndShell._encode(toy, 6, proc, [s])).data)



import threading
from enum import IntEnum
from typing import NamedTuple, Callable, Dict, List

from spherov2.commands.sphero import ReverseFlags, RollModes
from spherov2.controls import PacketDecodingException, CommandExecuteError
from spherov2.helper import packet_chk, to_bytes


class Packet:
    """Packet protocol v1.2, from https://docs.gosphero.com/api/Sphero_API_1.20.pdf"""

    SOP = 0xff
    ASYNC = 0xfe

    class Error(IntEnum):
        command_succeeded = 0x00
        non_specific_error = 0x01
        checksum_failure = 0x02
        command_fragment = 0x03
        unknown_command_id = 0x04
        unsupported_command = 0x05
        bad_message_format = 0x06
        parameter_value_invalid = 0x07
        failed_to_execute = 0x08
        unknown_device_id = 0x09
        voltage_too_low = 0x31
        illegal_page_number = 0x32
        flash_fail = 0x33
        main_application_corrupt = 0x34
        message_timeout = 0x35

    class Request(NamedTuple):
        """[SOP1, SOP2, DID, CID, SEQ, DLEN, <data>, CHK]"""

        did: int
        cid: int
        seq: int
        data: bytearray

        @property
        def id(self):
            return Packet.SOP, self.seq

        @property
        def dlen(self):
            return len(self.data) + 1

        def build(self) -> bytearray:
            payload = bytearray([Packet.SOP, Packet.SOP, self.did, self.cid, self.seq, self.dlen, *self.data])
            payload.append(packet_chk(payload[2:]))
            return payload

    class Response(NamedTuple):
        """[SOP1, SOP2, MRSP, SEQ, DLEN, <data>, CHK]"""

        mrsp: 'Packet.Error'
        seq: int
        data: bytearray

        @property
        def id(self):
            return Packet.SOP, self.seq

        @property
        def dlen(self):
            return len(self.data) + 1

        def build(self) -> bytearray:
            payload = bytearray([Packet.SOP, Packet.SOP, self.mrsp, self.seq, self.dlen, *self.data])
            payload.append(packet_chk(payload[2:]))
            return payload

        def check_error(self):
            if self.mrsp != Packet.Error.command_succeeded:
                raise CommandExecuteError(self.mrsp)

    class Async(NamedTuple):
        """[SOP1, SOP2, ID CODE, DLEN-MSB, DLEN-LSB, <data>, CHK]"""

        id_code: int
        data: bytearray

        @property
        def id(self):
            return Packet.ASYNC, self.id_code

        @property
        def dlen(self):
            return to_bytes(len(self.data) + 1, 2)

        def build(self) -> bytearray:
            payload = bytearray([Packet.SOP, Packet.ASYNC, self.id_code, *self.dlen, *self.data])
            payload.append(packet_chk(payload[2:]))
            return payload

    @staticmethod
    def parse_response(data) -> Response:
        chk = data.pop()
        if chk != packet_chk(data):
            raise PacketDecodingException('Bad response checksum')
        mrsp, seq, _, *data = data
        return Packet.Response(Packet.Error(mrsp), seq, bytearray(data))

    @staticmethod
    def parse_async(data) -> Async:
        chk = data.pop()
        if chk != packet_chk(data):
            raise PacketDecodingException('Bad response checksum')
        id_code, _, _, *data = data
        return Packet.Async(id_code, bytearray(data))

    class Manager:
        def __init__(self):
            self.__seq = 0

        def new_packet(self, did, cid, _, data=None):
            packet = Packet.Request(did, cid, self.__seq, bytearray(data or []))
            self.__seq = (self.__seq + 1) % 0x100
            return packet

    class Collector:
        def __init__(self, callback):
            self.__callback = callback
            self.__data = bytearray()

        def add(self, data):
            if not self.__data:
                while data and data[0] != Packet.SOP:
                    data.pop(0)
            self.__data.extend(data)
            while len(self.__data) > 4:
                sop1, sop2, *payload = self.__data
                if sop1 != Packet.SOP:
                    self.__data.clear()
                    raise PacketDecodingException('Unexpected start of packet')
                if sop2 == Packet.SOP:
                    _, _, dlen, *remain = payload
                    if dlen > len(remain):
                        break
                    self.__callback(Packet.parse_response(payload[:dlen + 3]))
                    self.__data = remain[dlen:]
                elif sop2 == Packet.ASYNC:
                    _, dlen_msb, dlen_lsb, *remain = payload
                    dlen = (dlen_msb << 8) | dlen_lsb
                    if dlen > len(remain):
                        break
                    self.__callback(Packet.parse_async(payload[:dlen + 3]))
                    self.__data = remain[dlen:]
                else:
                    raise PacketDecodingException('Unexpected start of packet 2')


class DriveControl:
    def __init__(self, toy):
        self.__toy = toy
        self.__is_aiming = False

    def roll_start(self, heading, speed):
        flag = ReverseFlags.OFF
        if speed < 0:
            flag = ReverseFlags.ON
            heading = (heading + 180) % 360
        speed = min(255, abs(speed))
        self.__toy.roll(speed, heading, RollModes.GO, flag)

    def roll_stop(self, heading):
        self.__toy.roll(0, heading, RollModes.STOP, ReverseFlags.OFF)

    def set_heading(self, heading):
        self.__toy.roll(0, heading, RollModes.CALIBRATE, ReverseFlags.OFF)

    def set_stabilization(self, stabilize):
        self.__toy.set_stabilization(stabilize)

    def set_raw_motors(self, left_mode, left_speed, right_mode, right_speed):
        self.__toy.set_raw_motors(left_mode, left_speed, right_mode, right_speed)


class FirmwareUpdateControl:
    def __init__(self, toy):
        self.__toy = toy


class LedControl:
    def __init__(self, toy):
        self.__toy = toy


class SensorControl:
    def __init__(self, toy):
        toy.add_sensor_streaming_data_notify_listener(self.__sensor_streaming_data)

        self.__toy = toy
        self.__count = 0
        self.__interval = 250
        self.__enabled = {}
        self.__enabled_extended = {}
        self.__listeners = set()

    def add_sensor_data_listener(self, listener: Callable[[Dict[str, Dict[str, float]]], None]):
        self.__listeners.add(listener)

    def remove_sensor_data_listener(self, listener: Callable[[Dict[str, Dict[str, float]]], None]):
        self.__listeners.remove(listener)

    def __sensor_streaming_data(self, sensor_data: List[int]):
        data = {}

        def __new_data():
            n = {}
            for name, component in components.items():
                d = sensor_data.pop(0)
                if component.modifier:
                    d = component.modifier(d)
                n[name] = d
            if self.__toy.name.startswith('2B') and sensor in ['locator', 'velocity']:
                n['x'], n['y'] = -n['y'], n['x']
            data[sensor] = n

        for sensor, components in self.__toy.sensors.items():
            if sensor in self.__enabled:
                __new_data()
        for sensor, components in self.__toy.extended_sensors.items():
            if sensor in self.__enabled_extended:
                __new_data()

        for f in self.__listeners:
            threading.Thread(target=f, args=(data,)).start()

    def set_count(self, count: int):
        if count >= 0 and count != self.__count:
            self.__count = count
            self.__update()

    def set_interval(self, interval: int):
        if interval >= 0 and interval != self.__interval:
            self.__interval = interval * 4 // 10
            if self.__interval == 0 and interval > 0:
                self.__interval = 1
            self.__update()

    def __update(self):
        sensors_mask = extended_sensors_mask = 0
        for sensor in self.__enabled.values():
            for component in sensor.values():
                sensors_mask |= component.bit
        for sensor in self.__enabled_extended.values():
            for component in sensor.values():
                extended_sensors_mask |= component.bit
        self.__toy.set_data_streaming(self.__interval, 1, sensors_mask, self.__count, extended_sensors_mask)

    def enable(self, *sensors):
        for sensor in sensors:
            if sensor in self.__toy.sensors:
                self.__enabled[sensor] = self.__toy.sensors[sensor]
            elif sensor in self.__toy.extended_sensors:
                self.__enabled_extended[sensor] = self.__toy.extended_sensors[sensor]
        self.__update()

    def disable(self, *sensors):
        for sensor in sensors:
            self.__enabled.pop(sensor, None)
            self.__enabled_extended.pop(sensor, None)
        self.__update()

    def disable_all(self):
        self.__enabled.clear()
        self.__enabled_extended.clear()
        self.__update()


class StatsControl:
    def __init__(self, toy):
        self.__toy = toy



from spherov2.commands.core import Core
from spherov2.toy.ollie import Ollie
from spherov2.types import ToyType


class BB8(Ollie):
    toy_type = ToyType('BB-8', 'BB-', 'BB', .06)

    # Core
    get_factory_config_block_crc = Core.get_factory_config_block_crc



from typing import NamedTuple


class Version(NamedTuple):
    major: int
    minor: int
    revision: int


class LastErrorInfo(NamedTuple):
    file_name: bytes
    line_number: int
    data: bytes


class ConfigBlock(NamedTuple):
    metadata_version: int
    config_block_version: int
    application_data: bytes


class ManufacturingDate(NamedTuple):
    year: int
    month: int
    day: int


class EventLogStatus(NamedTuple):
    log_capacity: int
    bytes_used: int
    events_in_log: int



import asyncio
import struct
import sys
from typing import Optional

import bleak

from spherov2.adapter.tcp_consts import RequestOp, ResponseOp
from spherov2.helper import to_bytes, to_int


async def process_connection(reader: asyncio.streams.StreamReader, writer: asyncio.streams.StreamWriter):
    peer = writer.get_extra_info('peername')

    def callback(char, d):
        if writer.is_closing():
            return
        char = char.encode('ascii')
        writer.write(ResponseOp.ON_DATA + to_bytes(len(char),
                     2) + char + to_bytes(len(d), 1) + d)
        asyncio.ensure_future(writer.drain())

    print('Incoming connection from %s:%d' % peer)
    adapter: Optional[bleak.BleakClient] = None

    try:
        while True:
            cmd = await reader.readexactly(1)
            if cmd == RequestOp.SCAN:
                timeout = struct.unpack('!f', await reader.readexactly(4))[0]
                try:
                    toys = await bleak.BleakScanner.discover(timeout)
                except BaseException as e:
                    err = str(e)[:0xffff].encode('utf_8')
                    writer.write(ResponseOp.ERROR +
                                 to_bytes(len(err), 2) + err)
                    await writer.drain()
                    continue
                writer.write(ResponseOp.OK + to_bytes(len(toys), 2))
                await writer.drain()
                for toy in toys:
                    name = toy.name.encode('utf_8')
                    addr = toy.address.encode('ascii')
                    writer.write(to_bytes(len(name), 2) + name +
                                 to_bytes(len(addr), 2) + addr)
                    await writer.drain()
            elif cmd == RequestOp.FIND:
                size = await reader.readexactly(2)
                name = (await reader.readexactly(size)).decode('utf-8')
                timeout = struct.unpack('!f', await reader.readexactly(4))[0]
                try:
                    toy = await bleak.BleakScanner.find_device_by_filter(lambda _, a: a.local_name == name, timeout)
                except BaseException as e:
                    err = str(e)[:0xffff].encode('utf_8')
                    writer.write(ResponseOp.ERROR +
                                 to_bytes(len(err), 2) + err)
                    await writer.drain()
                    continue
                writer.write(ResponseOp.OK)
                await writer.drain()
                name = toy.name.encode('utf_8')
                addr = toy.address.encode('ascii')
                writer.write(to_bytes(len(name), 2) + name +
                             to_bytes(len(addr), 2) + addr)
                await writer.drain()
            elif cmd == RequestOp.END:
                break
            else:
                seq_size = await reader.readexactly(3)
                seq, size = seq_size[0], to_int(seq_size[1:])
                data = (await reader.readexactly(size)).decode('ascii')
                try:
                    if cmd == RequestOp.INIT:
                        adapter = bleak.BleakClient(data, timeout=5.0)
                        await adapter.connect()
                    elif cmd == RequestOp.SET_CALLBACK:
                        await adapter.start_notify(data, callback)
                    elif cmd == RequestOp.WRITE:
                        size = to_int(await reader.readexactly(2))
                        payload = bytearray(await reader.readexactly(size))
                        await adapter.write_gatt_char(data, payload, True)
                except EOFError:
                    raise
                except BaseException as e:
                    err = str(e)[:0xffff].encode('utf_8')
                    writer.write(ResponseOp.ERROR +
                                 to_bytes(len(err), 2) + err + bytes([seq]))
                    await writer.drain()
                    continue
                writer.write(ResponseOp.OK + bytes([seq]))
                await writer.drain()
    finally:
        writer.close()
        if adapter and await adapter.is_connected():
            await adapter.disconnect()
        await writer.wait_closed()
        print('Disconnected from %s:%d' % peer)


if __name__ == '__main__':
    address = sys.argv[1] if len(sys.argv) > 1 else '0.0.0.0'
    port = int(sys.argv[2]) if len(sys.argv) > 2 else 50004
    loop = asyncio.get_event_loop()
    server = loop.run_until_complete(asyncio.start_server(
        process_connection, host=address, port=port))
    print('Server listening on %s:%d...' % (address, port))
    loop.run_until_complete(server.wait_closed())
